# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib.util
import json
import os
import shutil
import sys
import zipfile
from abc import ABC, abstractmethod
from types import MethodType
from typing import Any, Generic, Optional, Union

import numpy as np

import mlrun
from mlrun.artifacts import Artifact
from mlrun.execution import MLClientCtx
from mlrun.features import Feature

from .mlrun_interface import MLRunInterface
from .utils import CommonTypes, CommonUtils


class ModelHandler(ABC, Generic[CommonTypes.ModelType, CommonTypes.IOSampleType]):
    """
    An abstract interface for handling a model of the supported frameworks. The handler will support loading, saving
    and logging a model with all the required modules, custom objects and collected information about it.
    """

    # Framework name (Must be set when inheriting the class):
    FRAMEWORK_NAME = None  # type: str

    # Constant artifact names:
    _MODEL_FILE_ARTIFACT_NAME = "{}_model_file"
    _MODULES_MAP_ARTIFACT_NAME = "{}_modules_map.json"
    _CUSTOM_OBJECTS_MAP_ARTIFACT_NAME = "{}_custom_objects_map.json"
    _CUSTOM_OBJECTS_DIRECTORY_ARTIFACT_NAME = "{}_custom_objects.zip"

    # Constant defaults:
    _DEFAULT_ONNX_MODEL_NAME = "onnx_{}"

    def __init__(
        self,
        model: CommonTypes.ModelType = None,
        model_path: CommonTypes.PathType = None,
        model_name: Optional[str] = None,
        modules_map: Union[
            dict[str, Union[None, str, list[str]]], CommonTypes.PathType
        ] = None,
        custom_objects_map: Union[
            dict[str, Union[str, list[str]]], CommonTypes.PathType
        ] = None,
        custom_objects_directory: CommonTypes.PathType = None,
        context: MLClientCtx = None,
        **kwargs,
    ):
        """
        Initialize the handler. The model can be set here, so it won't require loading. Note you must provide at least
        one of `model` and `model_path`. If a model is not given, the files in the model path will be collected
        automatically to be ready for loading.

        :param model:                    Model to handle or None in case a loading parameters were supplied.
        :param model_path:               Path to the directory with the model files. Can be passed as a model object
                                         path in the following format:
                                         'store://models/<PROJECT_NAME>/<MODEL_NAME>:<VERSION>'
        :param model_name:               The model name for saving and logging the model:

                                         * Mandatory for loading the model from a local path.
                                         * If given a logged model (store model path) it will be read from the artifact.
                                         * If given a loaded model object and the model name is None, the name will be
                                           set to the model's object name / class.

        :param modules_map:              A dictionary of all the modules required for loading the model. Each key
                                         is a path to a module and its value is the object name to import from it. All
                                         the modules will be imported globally. If multiple objects needed to be
                                         imported from the same module a list can be given. The map can be passed as a
                                         path to a json file as well. For example:

                                         {
                                             "module1": None,  # => import module1
                                             "module2": ["func1", "func2"],  # => from module2 import func1, func2
                                             "module3.sub_module": "func3",  # => from module3.sub_module import func3
                                         }

                                         If the model path given is of a store object, the modules map will be read from
                                         the logged modules map artifact of the model.
        :param custom_objects_map:       A dictionary of all the custom objects required for loading the model. Each key
                                         is a path to a python file and its value is the custom object name to import
                                         from it. If multiple objects needed to be imported from the same py file a list
                                         can be given. The map can be passed as a path to a json file as well. For
                                         example:

                                         {
                                             "/.../custom_optimizer.py": "optimizer",
                                             "/.../custom_layers.py": ["layer1", "layer2"]
                                         }

                                         All the paths will be accessed from the given 'custom_objects_directory',
                                         meaning each py file will be read from 'custom_objects_directory/<MAP VALUE>'.
                                         If the model path given is of a store object, the custom objects map will be
                                         read from the logged custom object map artifact of the model.
                                         Notice: The custom objects will be imported in the order they came in this
                                         dictionary (or json). If a custom object is depended on another, make sure to
                                         put it below the one it relies on.
        :param custom_objects_directory: Path to the directory with all the python files required for the custom
                                         objects. Can be passed as a zip file as well (will be extracted during the run
                                         before loading the model). If the model path given is of a store object, the
                                         custom objects files will be read from the logged custom object artifact of the
                                         model.
        :param context:                  MLRun context to work with for logging the model.

        :raise MLRunInvalidArgumentError: In case one of the given parameters are invalid.
        """
        # Validate input:
        self._validate_model_parameters(model=model, model_path=model_path)
        self._validate_modules_parameter(modules_map=modules_map)
        self._validate_custom_objects_parameters(
            custom_objects_map=custom_objects_map,
            custom_objects_directory=custom_objects_directory,
        )

        # Set a default model name if needed - the class name of the given model:
        if model_name is None and model is not None:
            model_name = type(model).__name__

        # Store parameters:
        self._model_name = model_name
        self._model_path = model_path
        self._model = model
        self._modules_map = modules_map
        self._custom_objects_map = custom_objects_map
        self._custom_objects_directory = custom_objects_directory
        self._context = context

        # The imported modules from the map. None until the '_import_modules' method is called.
        self._modules = None  # type: Dict[str, Any]

        # The imported custom objects from the map. None until the '_import_custom_objects' method is called.
        self._custom_objects = None  # type: Dict[str, Any]

        # Local path to the model's primary file:
        self._model_file = kwargs.get("model_file", None)  # type: str

        # If the model path is of a store model object, this will be the 'ModelArtifact' object.
        self._model_artifact = kwargs.get("model_artifact", None)  # type: ModelArtifact

        # If the model path is of a store model object, this will be the extra data as DataItems ready to be downloaded.
        self._extra_data = kwargs.get("extra_data", {})  # type: Dict[str, CommonTypes.ExtraDataType]

        # If the model key is passed, override the default:
        self._model_key = kwargs.get("model_key", "model")

        # Setup additional properties for logging the model into a ModelArtifact:
        self._tag = ""
        self._inputs = None  # type: List[Feature]
        self._outputs = None  # type: List[Feature]
        self._labels = {}  # type: Dict[str, Union[str, int, float]]
        self._parameters = {}  # type: Dict[str, Union[str, int, float]]
        self._metrics = {}  # type: Dict[str, float]
        self._registered_artifacts = {}  # type: Dict[str, Artifact]

        # Set a flag to know if the user logged the model so its artifact is cached:
        self._is_logged = False

        # Collect the relevant files of the model into the handler (only in case the model was not provided):
        if model is None:
            self._collect_files()

    @property
    def model_name(self) -> str:
        """
        Get the handled model's name.

        :return: The handled model's name.
        """
        return self._model_name

    @property
    def model(self) -> CommonTypes.ModelType:
        """
        Get the handled model. Will return None in case the model is not initialized.

        :return: The handled model.
        """
        return self._model

    @property
    def model_file(self) -> str:
        """
        Get the model file path given to / located by this handler.

        :return: The model file path.
        """
        return self._model_file

    @property
    def context(self) -> mlrun.MLClientCtx:
        """
        Get the handler's MLRun context.

        :return: The handler's MLRun context.
        """
        return self._context

    @property
    def tag(self) -> str:
        """
        Get the model's tag.

        :return: The model's tag.
        """
        return self._tag

    @property
    def inputs(self) -> Union[list[Feature], None]:
        """
        Get the input ports features list of this model's artifact. If the inputs are not set, None will be returned.

        :return: The input ports features list if its set, otherwise None.
        """
        return self._inputs

    @property
    def outputs(self) -> Union[list[Feature], None]:
        """
        Get the output ports features list of this model's artifact. If the outputs are not set, None will be returned.

        :return: The output ports features list if its set, otherwise None.
        """
        return self._outputs

    @property
    def labels(self) -> dict[str, str]:
        """
        Get the labels dictionary of this model's artifact. These will be the labels that will be logged with the model.

        :return: The model's artifact labels.
        """
        return self._labels

    @property
    def parameters(self) -> dict[str, str]:
        """
        Get the parameters dictionary of this model's artifact. These will be the parameters that will be logged with
        the model.

        :return: The model's artifact parameters.
        """
        return self._parameters

    def get_artifacts(
        self, committed_only: bool = False
    ) -> dict[str, CommonTypes.ExtraDataType]:
        """
        Get the registered artifacts of this model's artifact. By default all the artifacts (logged and to be logged -
        committed only) will be returned. To get only the artifacts registered in the current run whom are committed and
        not logged yet, set the 'committed_only' flag to True.

        :param committed_only: Whether to return only the artifacts in queue to be logged or all the artifacts
                               registered to this model (logged and not logged).

        :return: The artifacts registered to this model.
        """
        if committed_only:
            return self._registered_artifacts
        return {**self._extra_data, **self._registered_artifacts}

    def set_model_name(self, model_name: str):
        """
        Set the handled model name. The 'save' and 'log' methods will use the new name for the files and logs. Keep in
        mind that changing the model's name before calling 'load' will fail as now the handler won't look for the
        correct files.

        :param model_name: The new model name to use.
        """
        self._model_name = model_name

    def set_context(self, context: MLClientCtx):
        """
        Set this handler MLRun context.

        :param context: The context to set to.
        """
        self._context = context

    def set_tag(self, tag: str):
        """
        Set the tag this model will be logged with.

        :param tag: The model tag to set.
        """
        self._tag = tag

    def set_inputs(
        self,
        from_sample: CommonTypes.IOSampleType = None,
        features: Optional[list[Feature]] = None,
        **kwargs,
    ):
        """
        Read the inputs property of this model to be logged along with it. The inputs can be set directly by passing the
        input features or to be read by a given input sample.

        :param from_sample: Read the inputs properties from a given input sample to the model.
        :param features:    List of MLRun.features.Feature to set.

        :raise MLRunInvalidArgumentError: If both parameters were passed.
        """
        # Validate parameters:
        if from_sample is not None and features is not None:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "The inputs can either be read from a given sample or from a features list. Both parameters cannot be "
                "passed."
            )

        # Set the inputs:
        self._inputs = (
            features
            if features is not None
            else self._read_io_samples(samples=from_sample)
        )

    def set_outputs(
        self,
        from_sample: CommonTypes.IOSampleType = None,
        features: Optional[list[Feature]] = None,
        **kwargs,
    ):
        """
        Read the outputs property of this model to be logged along with it. The outputs can be set directly by passing
        the output features or to be read by a given output sample.

        :param from_sample: Read the outputs properties from a given output sample from the model.
        :param features:    List of MLRun.features.Feature to set.

        :raise MLRunInvalidArgumentError: If both parameters were passed.
        """
        # Validate parameters:
        if from_sample is not None and features is not None:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "The outputs can either be read from a given sample or from a features list. Both parameters cannot be "
                "passed."
            )

        # Set the outputs:
        self._outputs = (
            features
            if features is not None
            else self._read_io_samples(samples=from_sample)
        )

    def set_labels(
        self,
        to_add: Optional[dict[str, Union[str, int, float]]] = None,
        to_remove: Optional[list[str]] = None,
    ):
        """
        Update the labels dictionary of this model artifact.

        :param to_add:    The labels to add.
        :param to_remove: A list of labels keys to remove.
        """
        # Update the labels:
        if to_add is not None:
            self._labels.update(to_add)

        # Remove labels:
        if to_remove is not None:
            for label in to_remove:
                self._labels.pop(label)

    def set_parameters(
        self,
        to_add: Optional[dict[str, Union[str, int, float]]] = None,
        to_remove: Optional[list[str]] = None,
    ):
        """
        Update the parameters dictionary of this model artifact.

        :param to_add:    The parameters to add.
        :param to_remove: A list of parameters keys to remove.
        """
        # Update the parameters:
        if to_add is not None:
            self._parameters.update(to_add)

        # Remove parameters:
        if to_remove is not None:
            for label in to_remove:
                self._parameters.pop(label)

    def set_metrics(
        self,
        to_add: Optional[dict[str, CommonTypes.ExtraDataType]] = None,
        to_remove: Optional[list[str]] = None,
    ):
        """
        Update the metrics dictionary of this model artifact.

        :param to_add:    The metrics to add.
        :param to_remove: A list of metrics keys to remove.
        """
        # Update the extra data:
        if to_add is not None:
            self._metrics.update(to_add)

        # Remove extra data:
        if to_remove is not None:
            for label in to_remove:
                self._metrics.pop(label)

    def set_extra_data(
        self,
        to_add: Optional[dict[str, CommonTypes.ExtraDataType]] = None,
        to_remove: Optional[list[str]] = None,
    ):
        """
        Update the extra data dictionary of this model artifact.

        :param to_add:    The extra data to add.
        :param to_remove: A list of extra data keys to remove.
        """
        # Update the extra data:
        if to_add is not None:
            self._extra_data.update(to_add)

        # Remove extra data:
        if to_remove is not None:
            for label in to_remove:
                self._extra_data.pop(label)

    def register_artifacts(
        self, artifacts: Union[Artifact, list[Artifact], dict[str, Artifact]]
    ):
        """
        Register the given artifacts, so they will be logged as extra data with the model of this handler. Notice: The
        artifacts will be logged only when either 'log' or 'update' are called.

        :param artifacts: The artifacts to register. Can be passed as a single artifact, a list of artifacts or an
                          artifacts dictionary. In case of single artifact or a list of artifacts, the artifacts key
                          will be used as its name in the extra data dictionary.
        """
        # If a single artifact is given, wrap in a list:
        if isinstance(artifacts, Artifact):
            artifacts = [artifacts]

        # If a list is given, prepare the artifacts dictionary:
        if isinstance(artifacts, list):
            artifacts = {artifact.key: artifact for artifact in artifacts}

        # Register the artifacts:
        self._registered_artifacts = {**self._registered_artifacts, **artifacts}

    @abstractmethod
    def save(
        self, output_path: CommonTypes.PathType = None, **kwargs
    ) -> Union[dict[str, Artifact], None]:
        """
        Save the handled model at the given output path.

        :param output_path:  The full path to the directory to save the handled model at. If not given, the context
                             stored will be used to save the model in the default artifacts location.

        :return The saved model artifacts dictionary if context is available and None otherwise.

        :raise MLRunRuntimeError:         In case there is no model initialized in this handler.
        :raise MLRunInvalidArgumentError: If an output path was not given, yet a context was not provided in
                                          initialization.
        """
        if self._model is None:
            raise mlrun.errors.MLRunRuntimeError(
                "Model cannot be save as it was not given in initialization or loaded during this run."
            )
        if output_path is None and self._context is None:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "An output path was not given and a context was not provided during the initialization of "
                "this model handler. To save the model, one of the two parameters must be supplied."
            )
        return None

    @abstractmethod
    def load(self, **kwargs):
        """
        Load the specified model in this handler. The custom objects will be imported right before loading the model. To
        access the model, call the 'model' property.
        """
        # If a model instance is already loaded, delete it from memory:
        if self._model:
            del self._model

        # Import the modules if needed (will be only imported once):
        if self._modules is None:
            self._import_modules()

        # Import the custom objects if needed (will be only imported once):
        if self._custom_objects is None:
            self._import_custom_objects()

    @abstractmethod
    def to_onnx(
        self, model_name: Optional[str] = None, optimize: bool = True, **kwargs
    ):
        """
        Convert the model in this handler to an ONNX model.

        :param model_name: The name to give to the converted ONNX model. If not given the default name will be the
                           stored model name with the suffix '_onnx'.
        :param optimize:   Whether to optimize the ONNX model using 'onnxoptimizer' before saving the model. Default:
                           True.

        :return: The converted ONNX model (onnx.ModelProto).
        """
        pass

    def log(
        self,
        tag: str = "",
        labels: Optional[dict[str, Union[str, int, float]]] = None,
        parameters: Optional[dict[str, Union[str, int, float]]] = None,
        inputs: Optional[list[Feature]] = None,
        outputs: Optional[list[Feature]] = None,
        metrics: Optional[dict[str, Union[int, float]]] = None,
        artifacts: Optional[dict[str, Artifact]] = None,
        extra_data: Optional[dict[str, CommonTypes.ExtraDataType]] = None,
        **kwargs,
    ):
        """
        Log the model held by this handler into the MLRun context provided. The stored values such as labels, parameters
        and artifacts will be used and updated where tag, inputs and outputs will be overridden if given here.

        :param tag:        Tag of a version to give to the logged model. Will override the stored tag in this handler.
        :param labels:     Labels to log the model with. Will be joined to the labels set.
        :param parameters: Parameters to log with the model. Will be joined to the parameters set.
        :param inputs:     A list of features this model expects to receive - the model's input ports. If already set,
                           will be overridden by the inputs given here.
        :param outputs:    A list of features this model expects to return - the model's output ports. If already set,
                           will be overridden by the outputs given here.
        :param metrics:    Metrics results to log with the model.
        :param artifacts:  Artifacts to log the model with. Will be joined to the registered artifacts and added to the
                           extra data.
        :param extra_data: Extra data to log with the model.

        :raise MLRunRuntimeError:         In case is no model in this handler.
        :raise MLRunInvalidArgumentError: If a context is missing.
        """
        # Validate there is a model and context:
        if self._model is None:
            raise mlrun.errors.MLRunRuntimeError(
                "Model cannot be logged as it was not given in initialization or loaded during this run."
            )
        if self._context is None:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "Cannot log model if a context was not provided during initialization."
            )

        # Save the model:
        model_artifacts = self.save()

        # Log the imported modules:
        modules_artifacts = self._log_modules() if self._modules_map is not None else {}

        # Log the custom objects:
        custom_objects_artifacts = (
            self._log_custom_objects() if self._custom_objects_map is not None else {}
        )

        # Read inputs and outputs ports:
        if inputs is not None:
            self.set_inputs(features=inputs)
        if outputs is not None:
            self.set_outputs(features=outputs)

        # Update the tag:
        if tag != "":
            self.set_tag(tag=tag)

        # Update labels, parameters and metrics:
        self.set_labels(to_add=labels)
        self.set_parameters(to_add=parameters)
        self.set_metrics(to_add=metrics)

        # Update the extra data:
        self._extra_data = {
            **self._extra_data,
            **(model_artifacts if model_artifacts is not None else {}),
            **self._registered_artifacts,
            **(artifacts if artifacts is not None else {}),
            **(extra_data if extra_data is not None else {}),
            **modules_artifacts,
            **custom_objects_artifacts,
        }
        self._registered_artifacts = {}

        # Log the model:
        self._model_artifact = self._context.log_model(
            key=self._model_key,
            db_key=self._model_name,
            model_file=self._model_file,
            tag=self._tag,
            inputs=self._inputs,
            outputs=self._outputs,
            framework=self.FRAMEWORK_NAME,
            labels=self._labels,
            parameters=self._parameters,
            metrics=self._metrics,
            extra_data={
                k: v
                for k, v in self._extra_data.items()
                if not isinstance(v, mlrun.DataItem)
            },
            algorithm=kwargs.get("algorithm", None),
            training_set=kwargs.get("sample_set", None),
            label_column=kwargs.get("target_columns", None),
            feature_vector=kwargs.get("feature_vector", None),
            feature_weights=kwargs.get("feature_weights", None),
        )

        # Mark the model is logged:
        self._is_logged = True

    def update(
        self,
        labels: Optional[dict[str, Union[str, int, float]]] = None,
        parameters: Optional[dict[str, Union[str, int, float]]] = None,
        inputs: Optional[list[Feature]] = None,
        outputs: Optional[list[Feature]] = None,
        metrics: Optional[dict[str, Union[int, float]]] = None,
        artifacts: Optional[dict[str, Artifact]] = None,
        extra_data: Optional[dict[str, CommonTypes.ExtraDataType]] = None,
        **kwargs,
    ):
        """
        Update the model held by this handler into the MLRun context provided, updating the model's artifact properties
        in the same model path provided.

        :param labels:     Labels to update or add to the model.
        :param parameters: Parameters to update or add to the model.
        :param inputs:     A list of features this model expects to receive - the model's input ports. If already set,
                           will be overridden by the inputs given here.
        :param outputs:    A list of features this model expects to return - the model's output ports. If already set,
                           will be overridden by the outputs given here.
        :param metrics:    Metrics results to log with the model.
        :param artifacts:  Artifacts to update or add to the model. Will be joined to the registered artifacts and added
                           to the extra data.
        :param extra_data: Extra data to update or add to the model.

        :raise MLRunInvalidArgumentError: In case a context is missing or the model path in this handler is missing or
                                          not of a store object.
        """
        # Validate model path:
        if self._model_artifact is None:
            if self._model_path is None:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    "Cannot update model if 'model_path' is not provided or if the model was never logged with this "
                    "handler."
                )
            elif not mlrun.datastore.is_store_uri(self._model_path):
                raise mlrun.errors.MLRunInvalidArgumentError(
                    "To update a model artifact the 'model_path' must be a store object."
                )

        # Read inputs and outputs ports:
        if inputs is not None:
            self.set_inputs(features=inputs)
        if outputs is not None:
            self.set_outputs(features=outputs)

        # Update labels and parameters:
        self.set_labels(to_add=labels)
        self.set_parameters(to_add=parameters)

        # Update the extra data:
        self._extra_data = {
            **self._extra_data,
            **self._registered_artifacts,
            **(artifacts if artifacts is not None else {}),
            **(extra_data if extra_data is not None else {}),
        }
        self._registered_artifacts = {}

        # Get the model artifact. If the model was logged during this run, use the artifact, otherwise use the
        # user's given model path:
        model_artifact = (
            self._context.get_artifact(self._model_name)
            if self._is_logged
            else self._model_path
        )

        # Update the model artifact:
        self._model_artifact = mlrun.artifacts.update_model(
            model_artifact=model_artifact,
            labels=self._labels,
            parameters=self._parameters,
            inputs=self._inputs,
            outputs=self._outputs,
            metrics=metrics,
            extra_data={
                k: v
                for k, v in self._extra_data.items()
                if not isinstance(v, mlrun.DataItem)
            },
            feature_vector=kwargs.get("feature_vector", None),
            feature_weights=kwargs.get("feature_weights", None),
            store_object=not self._is_logged,  # If the model was not logged, store the updated model in the database.
        )
        if self._is_logged:
            self._context.update_artifact(
                self._model_artifact
            )  # Update the cached model to the database.

    def _collect_files_from_store_object(self):
        """
        If the model path given is of a store object, collect the needed model files into this handler for later loading
        the model.
        """
        # Read the model artifact information:
        if len(self._model_artifact.inputs) > 0:
            self.set_inputs(features=list(self._model_artifact.inputs))
        if len(self._model_artifact.outputs) > 0:
            self.set_outputs(features=list(self._model_artifact.outputs))
        self.set_labels(to_add=self._model_artifact.labels)
        self.set_parameters(to_add=self._model_artifact.parameters)

        # Read the modules:
        if self._get_modules_map_artifact_name() in self._extra_data:
            self._modules_map = self._extra_data[
                self._get_modules_map_artifact_name()
            ].local()
        else:
            self._modules_map = None

        # Read the custom objects:
        if self._get_custom_objects_map_artifact_name() in self._extra_data:
            self._custom_objects_map = self._extra_data[
                self._get_custom_objects_map_artifact_name()
            ].local()
            self._custom_objects_directory = self._extra_data[
                self._get_custom_objects_directory_artifact_name()
            ].local()
        else:
            self._custom_objects_map = None
            self._custom_objects_directory = None

    @abstractmethod
    def _collect_files_from_local_path(self):
        """
        If the model path given is of a local path, search for the needed model files and collect them into this handler
        for later loading the model.
        """
        pass

    def _get_model_file_artifact_name(self) -> str:
        """
        Get the standard name for the model file artifact.

        :return: The model file artifact name.
        """
        return self._MODEL_FILE_ARTIFACT_NAME.format(self._model_name)

    def _get_modules_map_artifact_name(self) -> str:
        """
        Get the standard name for the modules map json artifact.

        :return: The modules map json artifact name.
        """
        return self._MODULES_MAP_ARTIFACT_NAME.format(self._model_name)

    def _get_custom_objects_map_artifact_name(self) -> str:
        """
        Get the standard name for the custom objects map json artifact.

        :return: The custom objects map json artifact name.
        """
        return self._CUSTOM_OBJECTS_MAP_ARTIFACT_NAME.format(self._model_name)

    def _get_custom_objects_directory_artifact_name(self) -> str:
        """
        Get the standard name for the custom objects directory zip artifact.

        :return: The custom objects directory zip artifact name.
        """
        return self._CUSTOM_OBJECTS_DIRECTORY_ARTIFACT_NAME.format(self._model_name)

    def _get_default_onnx_model_name(self, model_name: Union[str, None]):
        """
        Check if the given model name is None and if so will generate the default ONNX model name: 'onnx_<MODEL_NAME>'.

        :param model_name: The model name to check.

        :return: The given model name if its not None or the default ONNX model name.
        """
        return (
            self._DEFAULT_ONNX_MODEL_NAME.format(self._model_name)
            if model_name is None
            else model_name
        )

    def _collect_files(self):
        """
        Collect the files from the given model path.

        :raise MLRunInvalidArgumentError: In case the model path was not provided.
        """
        # Validate model path is set:
        if self._model_path is None:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "In order to collect the model's files a model path must be provided."
            )

        # Collect by the path's type:
        if mlrun.datastore.is_store_uri(self._model_path):
            # Check if the model object was already downloaded:
            if (
                self._model_file is None
                and self._model_artifact is None
                and self._extra_data == {}
            ):
                # Get the artifact and model file along with its extra data:
                (
                    self._model_file,
                    self._model_artifact,
                    self._extra_data,
                ) = mlrun.artifacts.get_model(self._model_path)
            # Check if the model name was not provided:
            if self._model_name is None:
                self._model_name = self._model_artifact.db_key
            # Continue to collect the files from the store object each framework requires:
            self._collect_files_from_store_object()
        else:
            if self._model_name is None:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    "The model name must be provided in the handler's initialization in order to collect the required "
                    "model files from a local path."
                )
            self._collect_files_from_local_path()

    def _import_modules(self):
        """
        Import the modules from the map provided.
        """
        # Initialize the custom objects dictionary:
        self._modules = {}

        # Check if modules parameters were provided:
        if self._modules_map is None:
            return

        # Read the modules map if given as a json:
        if isinstance(self._modules_map, str):
            with open(self._modules_map) as map_json_file:
                self._modules_map = json.loads(map_json_file.read())

        # Start importing the modules according to the map:
        for module_path, objects_names in self._modules_map.items():
            self._modules = {
                **self._modules,
                **self._import_module(
                    module_path=module_path,
                    objects_names=(
                        objects_names
                        if isinstance(objects_names, list) or objects_names is None
                        else [objects_names]
                    ),
                ),
            }

    def _import_custom_objects(self):
        """
        Import the custom objects from the map and directory provided.
        """
        # Initialize the custom objects dictionary:
        self._custom_objects = {}

        # Check if custom objects parameters were provided:
        if self._custom_objects_map is None:
            return

        # Read the custom objects map if given as a json:
        if isinstance(self._custom_objects_map, str):
            with open(self._custom_objects_map) as map_json_file:
                self._custom_objects_map = json.loads(map_json_file.read())

        # Unzip the custom objects files if the directory was given as a zip:
        if not os.path.isdir(self._custom_objects_directory):
            with zipfile.ZipFile(self._custom_objects_directory, "r") as zip_file:
                # Update the root directory of all the custom objects py files:
                self._custom_objects_directory = os.path.join(
                    os.path.dirname(self._custom_objects_directory),
                    os.path.basename(self._custom_objects_directory).rsplit(".", 1)[0],
                )
                # Extract the zip files into it:
                zip_file.extractall(self._custom_objects_directory)

        # Start importing the custom objects according to the map:
        for py_file, custom_objects_names in self._custom_objects_map.items():
            self._custom_objects = {
                **self._custom_objects,
                **self._import_custom_object(
                    py_file_path=os.path.abspath(
                        os.path.join(self._custom_objects_directory, py_file)
                    ),
                    objects_names=(
                        custom_objects_names
                        if isinstance(custom_objects_names, list)
                        else [custom_objects_names]
                    ),
                ),
            }

    def _log_modules(self) -> dict[str, Artifact]:
        """
        Log the modules, returning the modules map json file logged as an artifact.

        :return: The logged artifact in an 'extra data' style to be logged with the model.
        """
        # Initialize the returning artifacts dictionary:
        artifacts = {}

        # Create the custom objects map json file:
        modules_map_json = self._get_modules_map_artifact_name()
        if isinstance(self._modules_map, str):
            # The modules map is still a json path (model was not loaded but given as a live object):
            shutil.copy(self._modules_map, modules_map_json)
        else:
            # Dump the dictionary to json:
            with open(modules_map_json, "w") as json_file:
                json.dump(self._modules_map, json_file, indent=4)

        # Log the json file artifact:
        artifacts[modules_map_json] = self._context.log_artifact(
            modules_map_json,
            local_path=modules_map_json,
            artifact_path=self._context.artifact_path,
            db_key=False,
        )

        return artifacts

    def _log_custom_objects(self) -> dict[str, Artifact]:
        """
        Log the custom objects, returning their artifacts:

        * custom objects map json file - creating a json file from the custom objects map dictionary and logging it as
          an artifact.
        * custom objects directory zip - zipping the model's given custom objects directory along all of its content.

        :return: The logged artifacts in an 'extra data' style to be logged with the model.
        """
        # Initialize the returning artifacts dictionary:
        artifacts = {}

        # Create the custom objects map json file:
        custom_objects_map_json = self._get_custom_objects_map_artifact_name()
        if isinstance(self._custom_objects_map, str):
            # The custom objects map is still a json path (model was not loaded but given as a live object):
            shutil.copy(self._custom_objects_map, custom_objects_map_json)
        else:
            # Dump the dictionary to json:
            with open(custom_objects_map_json, "w") as json_file:
                json.dump(self._custom_objects_map, json_file, indent=4)

        # Log the json file artifact:
        artifacts[custom_objects_map_json] = self._context.log_artifact(
            custom_objects_map_json,
            local_path=custom_objects_map_json,
            artifact_path=self._context.artifact_path,
        )

        # Zip the custom objects directory:
        custom_objects_zip = self._get_custom_objects_directory_artifact_name()
        if self._custom_objects_directory.endswith(".zip"):
            # The custom objects are still zipped (model was not loaded but given as a live object):
            shutil.copy(self._custom_objects_directory, custom_objects_zip)
        else:
            # Zip the custom objects contents:
            shutil.make_archive(
                base_name=custom_objects_zip.rsplit(".", 1)[0],
                format="zip",
                root_dir=os.path.abspath(self._custom_objects_directory),
            )

        # Log the zip file artifact:
        artifacts[custom_objects_zip] = self._context.log_artifact(
            custom_objects_zip,
            local_path=custom_objects_zip,
            artifact_path=self._context.artifact_path,
        )

        return artifacts

    def _read_io_samples(
        self,
        samples: Union[CommonTypes.IOSampleType, list[CommonTypes.IOSampleType]],
    ) -> list[Feature]:
        """
        Read the given inputs / output sample to / from the model into a list of MLRun Features (ports) to log in
        the model's artifact.

        :param samples: The given inputs / output sample to / from the model.

        :return: The generated ports list.
        """
        # If there is only one input, wrap in a list:
        if not (isinstance(samples, list) or isinstance(samples, tuple)):
            samples = [samples]

        return [self._read_sample(sample=sample) for sample in samples]

    def _read_sample(self, sample: CommonTypes.IOSampleType) -> Feature:
        """
        Read the sample into a MLRun Feature. This abstract class is reading samples of 'numpy.ndarray'. For further
        types of samples, please inherit this method.

        :param sample: The sample to read.

        :return: The created Feature.

        :raise MLRunInvalidArgumentError: In case the given sample type cannot be read.
        """
        # Supported types:
        if isinstance(sample, np.ndarray):
            return Feature(
                value_type=CommonUtils.convert_np_dtype_to_value_type(
                    np_dtype=sample.dtype
                ),
                dims=list(sample.shape),
            )

        # Unsupported type:
        raise mlrun.errors.MLRunInvalidArgumentError(
            f"The sample type given '{type(sample)}' is not supported. The input / output ports are readable from "
            f"samples of the following types: np.ndarray"
        )

    @staticmethod
    def _validate_model_parameters(model_path: str, model: CommonTypes.ModelType):
        """
        Validate the given model parameters.

        :param model_path: Path to the directory with the model files. Can be passed as a model object path in the
                           following format: 'store://models/<PROJECT_NAME>/<MODEL_NAME>:<VERSION>'
        :param model:      Model to handle or None in case a loading parameters were supplied.

        :raise MLRunInvalidArgumentError: If both parameters were None or both parameters were provided.
        """
        if model_path is None and model is None:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "At least one of 'model' or 'model_path' must be provided to the model handler."
            )

    @staticmethod
    def _validate_modules_parameter(
        modules_map: Union[dict[str, Union[None, str, list[str]]], str],
    ):
        """
        Validate the given modules parameter.

        :param modules_map: A dictionary of all the modules required for loading the model. Each key is a path to a
                            module and its value is the object name to import from it. All the modules will be imported
                            globally. If multiple objects needed to be imported from the same module a list can be
                            given. The map can be passed as a path to a json file as well. For example:
                            {
                                "module1": None,  # => import module1
                                "module2": ["func1", "func2"],  # => from module2 import func1, func2
                                "module3.sub_module": "func3",  # => from module3.sub_module import func3
                            }
                            If the model path given is of a store object, the modules map will be read from the logged
                            modules map artifact of the model.

        :raise MLRunInvalidArgumentError: If the modules map is in incorrect file format or not exist.
        """
        if modules_map is not None:
            if isinstance(modules_map, str):
                if not (modules_map.endswith(".json") and os.path.exists(modules_map)):
                    raise mlrun.errors.MLRunInvalidArgumentError(
                        f"The 'modules_map' is either not found or not a path to a json file. Received: '{modules_map}'"
                    )

    @staticmethod
    def _validate_custom_objects_parameters(
        custom_objects_map: Union[dict[str, Union[str, list[str]]], str],
        custom_objects_directory: str,
    ):
        """
        Validate the given custom objects parameters.

        :param custom_objects_map:       A dictionary of all the custom objects required for loading the model. Each key
                                         is a path to a python file and its value is the custom object name to import
                                         from it. If multiple objects needed to be imported from the same py file a list
                                         can be given. The map can be passed as a path to a json file as well. For
                                         example:
                                         {
                                             "/.../custom_optimizer.py": "optimizer",
                                             "/.../custom_layers.py": ["layer1", "layer2"]
                                         }
                                         All the paths will be accessed from the given 'custom_objects_directory',
                                         meaning each py file will be read from 'custom_objects_directory/<MAP VALUE>'.
                                         Notice: The custom objects will be imported in the order they came in this
                                         dictionary (or json). If a custom object is depended on another, make sure to
                                         put it below the one it relies on.
        :param custom_objects_directory: Path to the directory with all the python files required for the custom
                                         objects. Can be passed as a zip file as well (will be extracted during the run
                                         before loading the model).

        :raise MLRunInvalidArgumentError: If the custom objects directory is given without the map or if the paths were
                                          in incorrect file formats or not exist.
        """
        # Validate that if one is provided (not None), both are provided:
        if (custom_objects_map is not None and custom_objects_directory is None) or (
            custom_objects_map is None and custom_objects_directory is not None
        ):
            raise mlrun.errors.MLRunInvalidArgumentError(
                "Either 'custom_objects_map' or 'custom_objects_directory' are None. Custom objects must be supplied "
                "with the custom object map dictionary (or json) and the directory (or zip) with all the python files."
            )

        # Validate that if the map is a path, it is a path to a json file:
        if custom_objects_map is not None:
            if isinstance(custom_objects_map, str):
                if not (
                    custom_objects_map.endswith(".json")
                    and os.path.exists(custom_objects_map)
                ):
                    raise mlrun.errors.MLRunInvalidArgumentError(
                        f"The 'custom_objects_map' is either not found or not a path to a json file. "
                        f"received: '{custom_objects_map}'"
                    )

        # Validate that the path is of a directory or a zip file:
        if custom_objects_directory is not None:
            if not (
                os.path.isdir(custom_objects_directory)
                or (
                    custom_objects_directory.endswith(".zip")
                    and os.path.exists(custom_objects_directory)
                )
            ):
                raise mlrun.errors.MLRunInvalidArgumentError(
                    f"The 'custom_objects_directory' is either not found or not a directory / zip file, "
                    f"received: '{custom_objects_directory}'"
                )

    @staticmethod
    def _import_module(
        module_path: str, objects_names: Union[list[str], None]
    ) -> dict[str, Any]:
        """
        Import the given objects by their names from the given module path by the following rules:

        * If 'objects_names' is None: import 'module_path'.
        * Otherwise: from 'module_path' import 'object_name'.

        If an object specified is already imported, a reference would simply be returned.

        :param module_path:   Path to the module with the objects to import.
        :param objects_names: The objects names to be imported from the given module.

        :return: The imported objects dictionary where the keys are the objects names and the values are their imported
                 objects.
        """
        # Initialize the imports dictionary:
        module_imports = {}

        # Check if the module is already imported:
        if module_path in sys.modules:
            # It is already imported:
            module = sys.modules[module_path]
        else:
            # Import the module:
            module = importlib.import_module(module_path)

        # Check what to import from the module:
        if objects_names is None:
            # Import the entire module (import X):
            module_imports[module_path] = module
        else:
            # Import multiple objects (from X import Y, Z, ...):
            module_imports = {
                object_name: getattr(module, object_name)
                for object_name in objects_names
            }

        # Update the globals dictionary with the module imports:
        globals().update(module_imports)

        return module_imports

    @staticmethod
    def _import_custom_object(
        py_file_path: str, objects_names: list[str]
    ) -> dict[str, Any]:
        """
        Import the given objects by their names from the given python file as: from 'py_file_path' import 'object_name'.
        If an object specified is already imported, a reference would simply be returned.

        :param py_file_path:  Path to the python file with the objects code.
        :param objects_names: The objects names to be imported from the given python file.

        :return: The imported objects dictionary where the keys are the objects names and the values are their imported
                 objects.
        """
        # Initialize the imports dictionary:
        objects_imports = {}

        # Go through the given classes:
        for object_name in objects_names:
            if object_name in sys.modules:
                # It is already imported:
                objects_imports[object_name] = sys.modules[object_name]
            else:
                # Import the custom object:
                spec = importlib.util.spec_from_file_location(
                    name=object_name, location=py_file_path
                )
                module = importlib.util.module_from_spec(spec=spec)
                spec.loader.exec_module(module)
                # Get the imported class and store it:
                objects_imports[object_name] = getattr(module, object_name)

        return objects_imports


def with_mlrun_interface(interface: type[MLRunInterface]):
    """
    Decorator configure for decorating a ModelHandler method (expecting 'self' to be the first argument) to add the
    given MLRun interface into the model before executing the method and remove it afterwards.

    :param interface: The MLRun interface to add.

    :return: The method decorator.
    """

    def decorator(model_handler_method: MethodType):
        def wrapper(model_handler: ModelHandler, *args, **kwargs):
            # Check if the interface is applied to the model inside the handler:
            is_applied = interface.is_applied(obj=model_handler.model)
            # If the interface is not applied, add it:
            if not is_applied:
                interface.add_interface(obj=model_handler.model)
            # Call the method:
            returned_value = model_handler_method(self=model_handler, *args, **kwargs)
            # If the interface was not applied, remove it:
            if not is_applied:
                interface.remove_interface(obj=model_handler.model)
            return returned_value

        return wrapper

    return decorator


def without_mlrun_interface(interface: type[MLRunInterface]):
    """
    Decorator configure for decorating a ModelHandler method (expecting 'self' to be the first argument) to remove the
    given MLRun interface from the model before executing the method and restore it afterwards.

    :param interface: The MLRun interface to remove.

    :return: The method decorator.
    """

    def decorator(model_handler_method: MethodType):
        def wrapper(model_handler: ModelHandler, *args, **kwargs):
            # Check if the interface is applied to the model inside the handler:
            is_applied = interface.is_applied(obj=model_handler.model)
            # If the interface is applied, remove it:
            restoration_information = None
            if is_applied:
                restoration_information = interface.remove_interface(
                    obj=model_handler.model
                )
            # Call the method:
            returned_value = model_handler_method(self=model_handler, *args, **kwargs)
            # If the interface was applied, add it:
            if is_applied:
                interface.add_interface(
                    obj=model_handler.model,
                    restoration=restoration_information,
                )
            return returned_value

        return wrapper

    return decorator



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union

import mlrun
from mlrun.features import Entity


def spark_session_update_hadoop_options(session, spark_options) -> dict[str, str]:
    hadoop_conf = session.sparkContext._jsc.hadoopConfiguration()
    non_hadoop_spark_options = {}

    for key, value in spark_options.items():
        if key.startswith("spark.hadoop."):
            key = key[len("spark.hadoop.") :]
            original_value = hadoop_conf.get(key, None)
            if original_value and original_value != value:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    f"The 'spark.hadoop.{key}' value is in conflict due to a discrepancy "
                    "with a previously established setting.\n"
                    f"This issue arises if 'spark.hadoop.{key}' has been preset in the Spark session, "
                    "or when using datastore profiles with differing security settings for this particular key."
                )
            hadoop_conf.set(key, value)
        else:
            non_hadoop_spark_options[key] = value
    return non_hadoop_spark_options


def check_special_columns_exists(
    spark_df, entities: list[Union[Entity, str]], timestamp_key: str, label_column: str
):
    columns = spark_df.columns
    entities = entities or []
    entities = [
        entity.name if isinstance(entity, Entity) else entity for entity in entities
    ]
    missing_entities = [entity for entity in entities if entity not in columns]
    cases_message = "Please check the letter cases (uppercase or lowercase)"
    if missing_entities:
        raise mlrun.errors.MLRunInvalidArgumentError(
            f"There are missing entities from dataframe during ingestion. missing_entities: {missing_entities}."
            f" {cases_message}"
        )
    if timestamp_key and timestamp_key not in columns:
        raise mlrun.errors.MLRunInvalidArgumentError(
            f"timestamp_key is missing from dataframe during ingestion. timestamp_key: {timestamp_key}."
            f" {cases_message}"
        )
    if label_column and label_column not in columns:
        raise mlrun.errors.MLRunInvalidArgumentError(
            f"label_column is missing from dataframe during ingestion. label_column: {label_column}. "
            f"{cases_message}"
        )



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable, Optional, Union

import numpy as np
from torch import Tensor
from torch.nn import Module, Parameter

import mlrun

from ..._common import LoggingMode
from ..._dl_common.loggers import Logger
from ..utils import PyTorchTypes
from .callback import Callback


class HyperparametersKeys:
    """
    For easy noting on which object to search for the hyperparameter to track with the logging callback in its
    initialization method.
    """

    MODEL = "model"
    TRAINING_SET = "training_set"
    VALIDATION_SET = "validation_set"
    LOSS_FUNCTION = "loss_function"
    OPTIMIZER = "optimizer"
    SCHEDULER = "scheduler"
    CUSTOM = "custom"


class LoggingCallback(Callback):
    """
    Callback for collecting data during training / evaluation. All the collected data will be available in this callback
    post the training / evaluation process and can be accessed via the 'training_results', 'validation_results',
    'static_hyperparameters', 'dynamic_hyperparameters' and 'summaries' properties.
    """

    class _MetricType:
        """
        Metric can be of two types, a loss metric or accuracy metric.
        """

        LOSS = "Loss"
        ACCURACY = "Accuracy"

    def __init__(
        self,
        context: mlrun.MLClientCtx = None,
        dynamic_hyperparameters: Optional[
            dict[
                str,
                tuple[
                    str,
                    Union[
                        list[Union[str, int]], Callable[[], PyTorchTypes.TrackableType]
                    ],
                ],
            ]
        ] = None,
        static_hyperparameters: Optional[
            dict[
                str,
                Union[PyTorchTypes.TrackableType, tuple[str, list[Union[str, int]]]],
            ]
        ] = None,
        auto_log: bool = False,
    ):
        """
        Initialize a logging callback with the given hyperparameters and logging configurations.

        :param context:                 MLRun context to automatically log its parameters if 'auto_log' is True.
        :param dynamic_hyperparameters: If needed to track a hyperparameter dynamically (sample it each epoch) it should
                                        be passed here. The parameter expects a dictionary where the keys are the
                                        hyperparameter chosen names and the values are tuples of object key and a list
                                        with the key chain. A key chain is a list of keys and indices to know how to
                                        access the needed hyperparameter. If the hyperparameter is not of accessible
                                        from any of the HyperparametersKeys, a custom callable method can be passed in
                                        the tuple instead of the key chain when providing the word
                                        HyperparametersKeys.CUSTOM. For example, to track the 'lr' attribute of
                                        an optimizer and a custom parameter, one should pass:
                                        {
                                            "learning rate": (HyperparametersKeys.OPTIMIZER, ["param_groups", 0, "lr"]),
                                            "custom parameter": (HyperparametersKeys.CUSTOM, get_custom_parameter)
                                        }
        :param static_hyperparameters:  If needed to track a hyperparameter one time per run it should be passed here.
                                        The parameter expects a dictionary where the keys are the
                                        hyperparameter chosen names and the values are the hyperparameter static value
                                        or a tuple of object key and a list with the key chain just like the dynamic
                                        hyperparameter. For example, to track the 'epochs' of an experiment run, one
                                        should pass:
                                        {
                                            "epochs": 7
                                        }
        :param auto_log:                Whether or not to enable auto logging, trying to track common static and dynamic
                                        hyperparameters.
        """
        super().__init__()

        # Store the configurations:
        self._dynamic_hyperparameters_keys = (
            dynamic_hyperparameters if dynamic_hyperparameters is not None else {}
        )
        self._static_hyperparameters_keys = (
            static_hyperparameters if static_hyperparameters is not None else {}
        )

        # Initialize the logger:
        self._logger = Logger(context=context)

        # Setup the logger flags:
        self._is_training = None  # type: bool
        self._auto_log = auto_log

    def get_training_results(self) -> dict[str, list[list[float]]]:
        """
        Get the training results logged. The results will be stored in a dictionary where each key is the metric name
        and the value is a list of lists of values. The first list is by epoch and the second list is by iteration
        (batch).

        :return: The training results.
        """
        return self._logger.training_results

    def get_validation_results(self) -> dict[str, list[list[float]]]:
        """
        Get the validation results logged. The results will be stored in a dictionary where each key is the metric name
        and the value is a list of lists of values. The first list is by epoch and the second list is by iteration
        (batch).

        :return: The validation results.
        """
        return self._logger.validation_results

    def get_static_hyperparameters(self) -> dict[str, PyTorchTypes.TrackableType]:
        """
        Get the static hyperparameters logged. The hyperparameters will be stored in a dictionary where each key is the
        hyperparameter name and the value is his logged value.

        :return: The static hyperparameters.
        """
        return self._logger.static_hyperparameters

    def get_dynamic_hyperparameters(
        self,
    ) -> dict[str, list[PyTorchTypes.TrackableType]]:
        """
        Get the dynamic hyperparameters logged. The hyperparameters will be stored in a dictionary where each key is the
        hyperparameter name and the value is a list of his logged values per epoch.

        :return: The dynamic hyperparameters.
        """
        return self._logger.dynamic_hyperparameters

    def get_summaries(self) -> dict[str, list[float]]:
        """
        Get the validation summaries of the metrics results. The summaries will be stored in a dictionary where each key
        is the metric names and the value is a list of all the summary values per epoch.

        :return: The validation summaries.
        """
        return self._logger.validation_summaries

    def get_epochs(self) -> int:
        """
        Get the overall epochs this callback participated in.

        :return: The overall epochs this callback participated in.
        """
        return self._logger.epochs

    def get_train_iterations(self) -> int:
        """
        Get the overall train iterations this callback participated in.

        :return: The overall train iterations this callback participated in.
        """
        return self._logger.training_iterations

    def get_validation_iterations(self) -> int:
        """
        Get the overall validation iterations this callback participated in.

        :return: The overall validation iterations this callback participated in.
        """
        return self._logger.validation_iterations

    def on_horovod_check(self, rank: int) -> bool:
        """
        Check whether this callback is fitting to run by the given horovod rank (worker).

        :param rank: The horovod rank (worker) id.

        :return: True if the callback is ok to run on this rank and false if not.
        """
        return rank == 0

    def on_run_begin(self):
        """
        After the run begins, this method will be called to setup the results and hyperparameters dictionaries for
        logging, noting the metrics names and logging the initial hyperparameters values (epoch 0).
        """
        # Set up the hyperparameters dictionaries:
        if self._auto_log:
            self._add_auto_hyperparameters()
        # # Static hyperparameters:
        for name, value in self._static_hyperparameters_keys.items():
            if isinstance(value, tuple):
                # Its a parameter that needed to be extracted via key chain.
                self._logger.log_static_hyperparameter(
                    parameter_name=name,
                    value=self._get_hyperparameter(source=value[0], key_chain=value[1]),
                )
            else:
                # Its a custom hyperparameter.
                self._logger.log_static_hyperparameter(parameter_name=name, value=value)
        # # Dynamic hyperparameters:
        for name, (source, key_chain) in self._dynamic_hyperparameters_keys.items():
            self._logger.log_dynamic_hyperparameter(
                parameter_name=name,
                value=self._get_hyperparameter(source=source, key_chain=key_chain),
            )

    def on_epoch_begin(self, epoch: int):
        """
        After the given epoch begins, this method will be called to append a new list to each of the metrics results for
        the new epoch.

        :param epoch: The epoch that is about to begin.
        """
        self._logger.log_epoch()

    def on_epoch_end(self, epoch: int):
        """
        Before the given epoch ends, this method will be called to log the dynamic hyperparameters as needed.

        :param epoch: The epoch that has just ended.
        """
        # Update the dynamic hyperparameters dictionary:
        if self._dynamic_hyperparameters_keys:
            for (
                parameter_name,
                (
                    source,
                    key_chain,
                ),
            ) in self._dynamic_hyperparameters_keys.items():
                self._logger.log_dynamic_hyperparameter(
                    parameter_name=parameter_name,
                    value=self._get_hyperparameter(source=source, key_chain=key_chain),
                )

    def on_train_begin(self):
        """
        After the training of the current epoch begins, this method will be called to set the mode for training.
        """
        self._is_training = True

    def on_train_end(self):
        """
        Before the training of the current epoch ends, this method will be called to lof the training summaries.
        """
        # Store the last training loss result of this epoch:
        loss_name = self._get_metric_name(
            metric_function=self._objects[self._ObjectKeys.LOSS_FUNCTION],
        )
        self._logger.log_training_summary(
            metric_name=loss_name,
            result=float(self._logger.training_results[loss_name][-1][-1]),
        )

        # Store the last training metrics results of this epoch:
        for metric_function in self._objects[self._ObjectKeys.METRIC_FUNCTIONS]:
            metric_name = self._get_metric_name(
                metric_function=metric_function,
            )
            self._logger.log_training_summary(
                metric_name=metric_name,
                result=float(self._logger.training_results[metric_name][-1][-1]),
            )

    def on_validation_begin(self):
        """
        After the validation (in a training case it will be per epoch) begins, this method will be called to set the
        mode for evaluation if the mode was not set for training already.
        """
        if self._is_training is None:
            self._is_training = False
            self._logger.set_mode(mode=LoggingMode.EVALUATION)

    def on_validation_end(
        self, loss_value: PyTorchTypes.MetricValueType, metric_values: list[float]
    ):
        """
        Before the validation (in a training case it will be per epoch) ends, this method will be called to log the
        validation results summaries.

        :param loss_value:    The loss summary of this validation.
        :param metric_values: The metrics summaries of this validation.
        """
        # Store the validation loss average of this epoch:
        self._logger.log_validation_summary(
            metric_name=self._get_metric_name(
                metric_function=self._objects[self._ObjectKeys.LOSS_FUNCTION],
            ),
            result=float(loss_value),
        )

        # Store the validation metrics averages of this epoch:
        for metric_function, metric_value in zip(
            self._objects[self._ObjectKeys.METRIC_FUNCTIONS], metric_values
        ):
            self._logger.log_validation_summary(
                metric_name=self._get_metric_name(
                    metric_function=metric_function,
                ),
                result=float(metric_value),
            )

    def on_train_batch_begin(self, batch: int, x: Tensor, y_true: Tensor):
        """
        After the training of the given batch begins, this method will be called to check whether this iteration needs
        to be logged.

        :param batch:  The current batch iteration of when this method is called.
        :param x:      The input part of the current batch.
        :param y_true: The true value part of the current batch.
        """
        self._logger.log_training_iteration()

    def on_validation_batch_begin(self, batch: int, x: Tensor, y_true: Tensor):
        """
        After the validation of the given batch begins, this method will be called to check whether this iteration needs
        to be logged.

        :param batch:  The current batch iteration of when this method is called.
        :param x:      The input part of the current batch.
        :param y_true: The true value part of the current batch.
        """
        self._logger.log_validation_iteration()

    def on_train_loss_end(self, loss_value: PyTorchTypes.MetricValueType):
        """
        After the training calculation of the loss, this method will be called to log the loss value.

        :param loss_value: The recent loss value calculated during training.
        """
        # Store the loss value at the current epoch:
        self._logger.log_training_result(
            metric_name=self._get_metric_name(
                metric_function=self._objects[self._ObjectKeys.LOSS_FUNCTION],
            ),
            result=float(loss_value),
        )

    def on_validation_loss_end(self, loss_value: PyTorchTypes.MetricValueType):
        """
        After the validating calculation of the loss, this method will be called to log the loss value.

        :param loss_value: The recent loss value calculated during validation.
        """
        # Store the loss value at the current epoch:
        self._logger.log_validation_result(
            metric_name=self._get_metric_name(
                metric_function=self._objects[self._ObjectKeys.LOSS_FUNCTION],
            ),
            result=float(loss_value),
        )

    def on_train_metrics_end(self, metric_values: list[PyTorchTypes.MetricValueType]):
        """
        After the training calculation of the metrics, this method will be called to log the metrics values.

        :param metric_values: The recent metric values calculated during training.
        """
        for metric_function, metric_value in zip(
            self._objects[self._ObjectKeys.METRIC_FUNCTIONS], metric_values
        ):
            self._logger.log_training_result(
                metric_name=self._get_metric_name(
                    metric_function=metric_function,
                ),
                result=float(metric_value),
            )

    def on_validation_metrics_end(
        self, metric_values: list[PyTorchTypes.MetricValueType]
    ):
        """
        After the validating calculation of the metrics, this method will be called to log the metrics values.

        :param metric_values: The recent metric values calculated during validation.
        """
        for metric_function, metric_value in zip(
            self._objects[self._ObjectKeys.METRIC_FUNCTIONS], metric_values
        ):
            self._logger.log_validation_result(
                metric_name=self._get_metric_name(
                    metric_function=metric_function,
                ),
                result=float(metric_value),
            )

    def _add_auto_hyperparameters(self):
        """
        Add auto log's hyperparameters if they are accessible. The automatic hyperparameters being added are:
        batch size, learning rate. In addition to that, the context parameters (if available) will be logged as static
        hyperparameters as well.
        """
        # Log the context parameters:
        if self._logger.context is not None:
            self._logger.log_context_parameters()

        # Add batch size:
        if "batch_size" not in self._static_hyperparameters_keys:
            if self._objects[self._ObjectKeys.TRAINING_SET] is not None and hasattr(
                self._objects[self._ObjectKeys.TRAINING_SET], "batch_size"
            ):
                self._static_hyperparameters_keys["batch_size"] = getattr(
                    self._objects[self._ObjectKeys.TRAINING_SET], "batch_size"
                )
            elif self._objects[self._ObjectKeys.VALIDATION_SET] is not None and hasattr(
                self._objects[self._ObjectKeys.VALIDATION_SET], "batch_size"
            ):
                self._static_hyperparameters_keys["batch_size"] = getattr(
                    self._objects[self._ObjectKeys.VALIDATION_SET], "batch_size"
                )

        # Add learning rate:
        learning_rate_key_chain = (
            HyperparametersKeys.OPTIMIZER,
            ["param_groups", 0, "lr"],
        )
        if (
            "lr" not in self._dynamic_hyperparameters_keys
            and "learning_rate" not in self._dynamic_hyperparameters_keys
        ):
            if self._objects[self._ObjectKeys.OPTIMIZER] is not None:
                try:
                    # Try to get the learning rate value:
                    self._get_hyperparameter(
                        source=learning_rate_key_chain[0],
                        key_chain=learning_rate_key_chain[1],
                    )
                    # If it passes without raising an exception, store the keychain ot it:
                    self._dynamic_hyperparameters_keys["lr"] = learning_rate_key_chain
                except (TypeError, KeyError, IndexError, ValueError):
                    pass

    def _get_hyperparameter(
        self,
        source: str,
        key_chain: Union[
            list[Union[str, int]], Callable[[], PyTorchTypes.TrackableType]
        ],
    ) -> PyTorchTypes.TrackableType:
        """
        Access the hyperparameter from the source using the given key chain.

        :param source:    The object string (out of 'HyperparametersKeys') to get the hyperparamter value from.
        :param key_chain: The keys and indices to get to the hyperparameter from the given source object or if the
                          source is equal to 'HyperparametersKeys.CUSTOM', the callable custom method to get the value
                          to track.

        :return: The hyperparameter value.

        :raise TypeError:  In case the source is 'HyperparametersKeys.CUSTOM' but the given value is not callable.
        :raise KeyError:   In case the one of the keys in the key chain is incorrect.
        :raise IndexError: In case the one of the keys in the key chain is incorrect.
        :raise MLRunInvalidArgumentError: In case the value is not trackable.
        """
        # Get the value using the provided key chain:
        if source == HyperparametersKeys.CUSTOM:
            # It is a custom callable method:
            try:
                value = key_chain()
            except TypeError:
                raise TypeError(
                    f"The given value of the source '{source}' "
                    f"is of type '{type(key_chain)}' and is not callable."
                )
        else:
            # Needed to be extracted via key chain:
            value = self._objects[source]
            for key in key_chain:
                try:
                    if isinstance(value, (dict, list, tuple)):
                        value = value[key]
                    else:
                        value = getattr(value, key)
                except (KeyError, IndexError, AttributeError) as KeyChainError:
                    raise KeyChainError(
                        f"Error during getting a hyperparameter value from the {source.__class__} object. The "
                        f"{value.__class__} in it does not have the following key/index from the key provided: {key}"
                    )

        # Parse the value:
        if isinstance(value, (Tensor, Parameter)):
            if value.numel() == 1:
                value = float(value)
            else:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    f"The parameter with the following key chain: {key_chain} is a pytorch.Tensor with {value.numel()} "
                    f"elements. PyTorch tensors are trackable only if they have 1 element."
                )
        elif isinstance(value, np.ndarray):
            if value.size == 1:
                value = float(value)
            else:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    f"The parameter with the following key chain: {key_chain} is a numpy.ndarray with {value.size} "
                    f"elements. numpy arrays are trackable only if they have 1 element."
                )
        elif not isinstance(value, (float, int, str, bool)):
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"The parameter with the following key chain: {key_chain} is of type '{type(value)}'. "
                f"The only trackable types are: float, int, str and bool."
            )
        return value

    @staticmethod
    def _get_metric_name(metric_function: PyTorchTypes.MetricFunctionType):
        """
        Get the given metric name.

        :param metric_function: The metric function to get its name.

        :return: The metric name.
        """
        if isinstance(metric_function, Module):
            function_name = metric_function.__class__.__name__
        else:
            function_name = metric_function.__name__
        return function_name



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Union

from pydantic.v1 import BaseModel

import mlrun.common.types


class GrafanaColumnType(mlrun.common.types.StrEnum):
    NUMBER = "number"
    STRING = "string"


class GrafanaColumn(BaseModel):
    text: str
    type: str


class GrafanaNumberColumn(GrafanaColumn):
    type: str = GrafanaColumnType.NUMBER


class GrafanaStringColumn(GrafanaColumn):
    type: str = GrafanaColumnType.STRING


class GrafanaTable(BaseModel):
    columns: list[GrafanaColumn]
    rows: list[list[Optional[Union[float, int, str]]]] = []
    type: str = "table"

    def add_row(self, *args):
        self.rows.append(list(args))


class GrafanaModelEndpointsTable(GrafanaTable):
    def __init__(self):
        columns = self._init_columns()
        super().__init__(columns=columns)

    @staticmethod
    def _init_columns():
        return [
            GrafanaColumn(text="endpoint_id", type=GrafanaColumnType.STRING),
            GrafanaColumn(text="endpoint_name", type=GrafanaColumnType.STRING),
            GrafanaColumn(text="endpoint_function", type=GrafanaColumnType.STRING),
            GrafanaColumn(text="endpoint_model", type=GrafanaColumnType.STRING),
            GrafanaColumn(text="endpoint_model_class", type=GrafanaColumnType.STRING),
            GrafanaColumn(text="error_count", type=GrafanaColumnType.NUMBER),
            GrafanaColumn(text="drift_status", type=GrafanaColumnType.NUMBER),
            GrafanaColumn(text="sampling_percentage", type=GrafanaColumnType.NUMBER),
        ]



# Copyright 2025 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime

import mlrun.common.schemas.model_monitoring as mm_schemas
import mlrun.feature_store.steps
from mlrun.utils import logger


class ProcessBeforeTDEngine(mlrun.feature_store.steps.MapClass):
    def __init__(self, **kwargs):
        """
        Process the data before writing to TDEngine. This step create the table name.

        :returns: Event as a dictionary which will be written into the TDEngine Metrics/Results tables.
        """
        super().__init__(**kwargs)

    def do(self, event):
        logger.info("Process event before writing to TDEngine", event=event)
        kind = event.get("kind")
        table_name = (
            f"{event[mm_schemas.WriterEvent.ENDPOINT_ID]}_"
            f"{event[mm_schemas.WriterEvent.APPLICATION_NAME]}"
        )
        if kind == mm_schemas.WriterEventKind.RESULT:
            # Write a new result
            event[mm_schemas.EventFieldType.TABLE_COLUMN] = (
                f"{table_name}_{event[mm_schemas.ResultData.RESULT_NAME]}"
            ).replace("-", "_")
        elif kind == mm_schemas.WriterEventKind.METRIC:
            # Write a new metric
            event[mm_schemas.EventFieldType.TABLE_COLUMN] = (
                f"{table_name}_{event[mm_schemas.MetricData.METRIC_NAME]}"
            ).replace("-", "_")
        event[mm_schemas.WriterEvent.START_INFER_TIME] = datetime.fromisoformat(
            event[mm_schemas.WriterEvent.START_INFER_TIME]
        )
        return event



# Copyright 2024 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
from typing import Optional
from urllib.parse import urlparse

import fsspec

from mlrun.datastore.base import DataStore


class HdfsStore(DataStore):
    def __init__(
        self, parent, schema, name, endpoint="", secrets: Optional[dict] = None
    ):
        super().__init__(parent, name, schema, endpoint, secrets)

        self.host = self._get_secret_or_env("HDFS_HOST")
        self.port = self._get_secret_or_env("HDFS_PORT")
        self.http_port = self._get_secret_or_env("HDFS_HTTP_PORT")
        self.user = self._get_secret_or_env("HDFS_USER")
        if not self.user:
            self.user = os.environ.get("HADOOP_USER_NAME", os.environ.get("USER"))

        self._filesystem = None

    @property
    def filesystem(self):
        if not self._filesystem:
            self._filesystem = fsspec.filesystem(
                "webhdfs",
                host=self.host,
                port=self.http_port,
                user=self.user,
            )
        return self._filesystem

    @property
    def url(self):
        return f"webhdfs://{self.host}:{self.http_port}"

    @property
    def spark_url(self):
        return f"hdfs://{self.host}:{self.port}"

    def rm(self, url, recursive=False, maxdepth=None):
        path = urlparse(url).path
        self.filesystem.rm(path=path, recursive=recursive, maxdepth=maxdepth)



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import abc
from typing import Optional

import IPython.display

import mlrun.common.constants as mlrun_constants
import mlrun.errors
import mlrun.launcher.base as launcher
import mlrun.lists
import mlrun.model
import mlrun.runtimes
import mlrun.utils
import mlrun.utils.version


class ClientBaseLauncher(launcher.BaseLauncher, abc.ABC):
    """
    Abstract class for common code between client launchers
    """

    def enrich_runtime(
        self,
        runtime: "mlrun.runtimes.base.BaseRuntime",
        project_name: Optional[str] = "",
        full: bool = True,
        client_version: str = "",
    ):
        runtime.try_auto_mount_based_on_config()
        runtime._fill_credentials()
        if project_name:
            runtime.metadata.project = project_name

    @staticmethod
    def prepare_image_for_deploy(runtime: "mlrun.runtimes.BaseRuntime"):
        """
        Check if the runtime requires to build the image.
        If build is needed, set the image as the base_image for the build.
        If image is not given set the default one.
        """
        if runtime.kind in mlrun.runtimes.RuntimeKinds.pure_nuclio_deployed_runtimes():
            return

        require_build = runtime.requires_build()
        image = runtime.spec.image
        # we allow users to not set an image, in that case we'll use the default
        if (
            not image
            and runtime.kind in mlrun.mlconf.function_defaults.image_by_kind.to_dict()
        ):
            image = mlrun.mlconf.function_defaults.image_by_kind.to_dict()[runtime.kind]

        mlrun.utils.helpers.warn_on_deprecated_image(image)

        # TODO: need a better way to decide whether a function requires a build
        if require_build and image and not runtime.spec.build.base_image:
            # when the function require build use the image as the base_image for the build
            runtime.spec.build.base_image = image
            runtime.spec.image = ""

    @staticmethod
    def _store_function(
        runtime: "mlrun.runtimes.BaseRuntime", run: "mlrun.run.RunObject"
    ):
        run.metadata.labels[mlrun_constants.MLRunInternalLabels.kind] = runtime.kind
        mlrun.runtimes.utils.enrich_run_labels(
            run.metadata.labels, [mlrun_constants.MLRunInternalLabels.owner]
        )
        if run.spec.output_path:
            run.spec.output_path = run.spec.output_path.replace(
                "{{run.user}}",
                run.metadata.labels[mlrun_constants.MLRunInternalLabels.owner],
            )
        db = runtime._get_db()
        if db and runtime.kind != "handler":
            struct = runtime.to_dict()
            hash_key = db.store_function(
                struct, runtime.metadata.name, runtime.metadata.project, versioned=True
            )
            run.spec.function = runtime._function_uri(hash_key=hash_key)

    @staticmethod
    def _refresh_function_metadata(runtime: "mlrun.runtimes.BaseRuntime"):
        try:
            meta = runtime.metadata
            db = runtime._get_db()
            db_func = db.get_function(meta.name, meta.project, meta.tag)
            if db_func and "status" in db_func:
                runtime.status = db_func["status"]
                if (
                    runtime.status.state
                    and runtime.status.state == "ready"
                    and runtime.kind
                    # We don't want to override the nuclio image here because the build happens in nuclio
                    # TODO: have a better way to check if nuclio function deploy started
                    and not hasattr(runtime.status, "nuclio_name")
                ):
                    runtime.spec.image = mlrun.utils.get_in(
                        db_func, "spec.image", runtime.spec.image
                    )
        except mlrun.errors.MLRunNotFoundError:
            pass

    @staticmethod
    def _log_track_results(is_child: bool, result: dict, run: "mlrun.run.RunObject"):
        """
        log commands to track results
        in jupyter, displays a table widget with the result
        else, logs CLI commands to track results and a link to the results in UI

        :param: runtime: A bool to determine whether runtime is child or not
        :param result:   Run result dict
        :param run:      Run object
        """
        uid = run.metadata.uid
        project = run.metadata.project

        # show ipython/jupyter result table widget
        results_tbl = mlrun.lists.RunList()
        if result:
            results_tbl.append(result)
        else:
            mlrun.utils.logger.info("no returned result (job may still be in progress)")
            results_tbl.append(run.to_dict())

        if mlrun.utils.is_jupyter and mlrun.mlconf.ipython_widget:
            results_tbl.show()
            print()
            ui_url = mlrun.utils.get_run_url(project, uid=uid, name=run.metadata.name)
            if ui_url:
                ui_url = f' or <a href="{ui_url}" target="_blank">click here</a> to open in UI'
            IPython.display.display(
                IPython.display.HTML(
                    f"<b> > to track results use the .show() or .logs() methods {ui_url}</b>"
                )
            )
        elif is_child:
            # TODO: Log sdk commands to track results instead of CLI commands
            project_flag = f"-p {project}" if project else ""
            info_cmd = f"mlrun get run {uid} {project_flag}"
            logs_cmd = f"mlrun logs {uid} {project_flag}"
            mlrun.utils.logger.info(
                "To track results use the CLI", info_cmd=info_cmd, logs_cmd=logs_cmd
            )
            ui_url = mlrun.utils.get_run_url(project, uid=uid, name=run.metadata.name)
            if ui_url:
                mlrun.utils.logger.info("Or click for UI", ui_url=ui_url)



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .alert import (
    AlertActivation,
    AlertActivations,
    AlertActiveState,
    AlertConfig,
    AlertCriteria,
    AlertNotification,
    AlertTemplate,
    Event,
)
from .api_gateway import (
    APIGateway,
    APIGatewayAuthenticationMode,
    APIGatewayBasicAuth,
    APIGatewayMetadata,
    APIGatewaysOutput,
    APIGatewaySpec,
    APIGatewayState,
    APIGatewayStatus,
    APIGatewayUpstream,
)
from .artifact import (
    Artifact,
    ArtifactCategories,
    ArtifactIdentifier,
    ArtifactMetadata,
    ArtifactSpec,
)
from .auth import (
    AuthInfo,
    AuthorizationAction,
    AuthorizationResourceTypes,
    AuthorizationVerificationInput,
    Credentials,
    ProjectsRole,
)
from .background_task import (
    BackgroundTask,
    BackgroundTaskList,
    BackgroundTaskMetadata,
    BackgroundTaskSpec,
    BackgroundTaskState,
    BackgroundTaskStatus,
)
from .client_spec import ClientSpec
from .clusterization_spec import (
    ClusterizationSpec,
    WaitForChiefToReachOnlineStateFeatureFlag,
)
from .common import ImageBuilder
from .constants import (
    APIStates,
    ArtifactPartitionByField,
    ClusterizationRole,
    DeletionStrategy,
    FeatureStorePartitionByField,
    HeaderNames,
    LogsCollectorMode,
    OrderType,
    PatchMode,
    RunPartitionByField,
    SortField,
)
from .datastore_profile import DatastoreProfile
from .events import (
    AuthSecretEventActions,
    EventClientKinds,
    EventsModes,
    SecretEventActions,
)
from .feature_store import (
    EntitiesOutput,
    EntitiesOutputV2,
    Entity,
    EntityListOutput,
    EntityRecord,
    Feature,
    FeatureListOutput,
    FeatureRecord,
    FeatureSet,
    FeatureSetDigestOutput,
    FeatureSetDigestOutputV2,
    FeatureSetDigestSpec,
    FeatureSetDigestSpecV2,
    FeatureSetIngestInput,
    FeatureSetIngestOutput,
    FeatureSetRecord,
    FeatureSetsOutput,
    FeatureSetSpec,
    FeatureSetsTagsOutput,
    FeaturesOutput,
    FeaturesOutputV2,
    FeatureVector,
    FeatureVectorRecord,
    FeatureVectorsOutput,
    FeatureVectorsTagsOutput,
)
from .frontend_spec import (
    ArtifactLimits,
    AuthenticationFeatureFlag,
    FeatureFlags,
    FrontendSpec,
    NuclioStreamsFeatureFlag,
    PreemptionNodesFeatureFlag,
    ProjectMembershipFeatureFlag,
)
from .function import FunctionState, PreemptionModes, SecurityContextEnrichmentModes
from .http import HTTPSessionRetryMode
from .hub import (
    HubCatalog,
    HubItem,
    HubObjectMetadata,
    HubSource,
    HubSourceSpec,
    IndexedHubSource,
    last_source_index,
)
from .k8s import NodeSelectorOperator, Resources, ResourceSpec
from .memory_reports import MostCommonObjectTypesReport, ObjectTypeReport
from .model_monitoring import (
    DriftStatus,
    EndpointMode,
    EndpointType,
    EndpointUID,
    EventFieldType,
    EventKeyMetrics,
    Features,
    FeatureSetFeatures,
    FeatureValues,
    FileTargetKind,
    GrafanaColumn,
    GrafanaNumberColumn,
    GrafanaStringColumn,
    GrafanaTable,
    ModelEndpoint,
    ModelEndpointCreationStrategy,
    ModelEndpointDriftValues,
    ModelEndpointList,
    ModelEndpointMetadata,
    ModelEndpointSchema,
    ModelEndpointSpec,
    ModelEndpointStatus,
    ModelMonitoringInfraLabel,
    ModelMonitoringMode,
    MonitoringFunctionNames,
    TSDBTarget,
    V3IOTSDBTables,
)
from .notification import (
    Notification,
    NotificationKind,
    NotificationSeverity,
    NotificationState,
    NotificationStatus,
    NotificationSummary,
    SetNotificationRequest,
)
from .object import ObjectKind, ObjectMetadata, ObjectSpec, ObjectStatus
from .pagination import PaginationInfo
from .partition import PartitionInterval
from .pipeline import PipelinesOutput, PipelinesPagination
from .project import (
    IguazioProject,
    Project,
    ProjectDesiredState,
    ProjectMetadata,
    ProjectOut,
    ProjectOutput,
    ProjectOwner,
    ProjectsOutput,
    ProjectSpec,
    ProjectSpecOut,
    ProjectState,
    ProjectStatus,
    ProjectSummariesOutput,
    ProjectSummary,
)
from .regex import RegexMatchModes
from .runs import RunIdentifier
from .runtime_resource import (
    GroupedByJobRuntimeResourcesOutput,
    GroupedByProjectRuntimeResourcesOutput,
    KindRuntimeResources,
    ListRuntimeResourcesGroupByField,
    RuntimeResource,
    RuntimeResources,
    RuntimeResourcesOutput,
)
from .schedule import (
    ScheduleCronTrigger,
    ScheduleIdentifier,
    ScheduleInput,
    ScheduleKinds,
    ScheduleOutput,
    ScheduleRecord,
    SchedulesOutput,
    ScheduleUpdate,
)
from .secret import (
    AuthSecretData,
    SecretKeysData,
    SecretProviderName,
    SecretsData,
    UserSecretCreationRequest,
)
from .serving import ModelRunnerStepData, ModelsData, MonitoringData
from .tag import Tag, TagObjects
from .workflow import (
    GetWorkflowResponse,
    RerunWorkflowRequest,
    WorkflowRequest,
    WorkflowResponse,
    WorkflowSpec,
)



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC

from .._common.artifacts_library import ArtifactsLibrary, Plan
from .plans import (
    CalibrationCurvePlan,
    ConfusionMatrixPlan,
    DatasetPlan,
    FeatureImportancePlan,
    ROCCurvePlan,
)
from .utils import MLTypes, MLUtils


class MLArtifactsLibrary(ArtifactsLibrary, ABC):
    """
    An abstract class for a ML framework artifacts library. Each ML framework should have an artifacts library for
    knowing what artifacts can be produced and their configurations. The default method must be implemented for when the
    user do not pass any plans.

    To add a plan to the library, simply write its name in the library class as a class variable pointing to the plan's
    class 'init_artifact' class method:

    some_artifact = SomeArtifactPlan
    """

    calibration_curve = CalibrationCurvePlan
    confusion_matrix = ConfusionMatrixPlan
    dataset = DatasetPlan
    feature_importance = FeatureImportancePlan
    roc_curve = ROCCurvePlan

    @classmethod
    def default(
        cls, model: MLTypes.ModelType, y: MLTypes.DatasetType = None, *args, **kwargs
    ) -> list[Plan]:
        """
        Get the default artifacts plans list of this framework's library.

        :param model: The model to check if its a regression model or a classification model.
        :param y:     The ground truth values to check if its multiclass and / or multi output.

        :return: The default artifacts plans list.
        """
        # Discover the algorithm functionality of the provided model:
        algorithm_functionality = MLUtils.get_algorithm_functionality(model=model, y=y)

        # Initialize the plans list:
        plans = [DatasetPlan(purpose=DatasetPlan.Purposes.TEST)]

        # Add classification plans:
        if algorithm_functionality.is_classification():
            if algorithm_functionality.is_single_output():
                plans += [
                    FeatureImportancePlan(),
                    ConfusionMatrixPlan(),
                    ROCCurvePlan(),
                ]
                if algorithm_functionality.is_binary_classification():
                    plans += [CalibrationCurvePlan()]

        # Add regression plans:
        if algorithm_functionality.is_regression():
            if algorithm_functionality.is_single_output():
                plans += [FeatureImportancePlan()]

        # Filter out the plans by probabilities requirement:
        if not hasattr(model, "predict_proba"):
            plans = [plan for plan in plans if not plan.need_probabilities]

        return plans



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mergedeep

import mlrun.common.types
import mlrun.errors


class PatchMode(mlrun.common.types.StrEnum):
    replace = "replace"
    additive = "additive"

    def to_mergedeep_strategy(self) -> mergedeep.Strategy:
        if self.value == PatchMode.replace:
            return mergedeep.Strategy.REPLACE
        elif self.value == PatchMode.additive:
            return mergedeep.Strategy.ADDITIVE
        else:
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"Unknown patch mode: {self.value}"
            )


class DeletionStrategy(mlrun.common.types.StrEnum):
    restrict = "restrict"
    restricted = "restricted"
    cascade = "cascade"
    cascading = "cascading"
    check = "check"

    @staticmethod
    def default():
        return DeletionStrategy.restricted

    def is_restricted(self):
        if self.value in [DeletionStrategy.restrict, DeletionStrategy.restricted]:
            return True
        return False

    def is_cascading(self):
        if self.value in [DeletionStrategy.cascade, DeletionStrategy.cascading]:
            return True
        return False

    def strategy_to_check(self):
        if self.value in [
            DeletionStrategy.check,
            DeletionStrategy.restrict,
            DeletionStrategy.restricted,
        ]:
            return True
        return False

    def to_nuclio_deletion_strategy(self) -> str:
        if self.is_restricted():
            return "restricted"
        elif self.is_cascading():
            return "cascading"
        elif self.value == DeletionStrategy.check.value:
            return "check"
        else:
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"Unknown deletion strategy: {self.value}"
            )

    def to_iguazio_deletion_strategy(self) -> str:
        if self.is_restricted():
            return "restricted"
        elif self.is_cascading():
            return "cascading"
        elif self.value == DeletionStrategy.check.value:
            raise NotImplementedError(
                "Iguazio does not support the check deletion strategy"
            )
        else:
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"Unknown deletion strategy: {self.value}"
            )


headers_prefix = "x-mlrun-"


class HeaderNames:
    projects_role = "x-projects-role"
    patch_mode = f"{headers_prefix}patch-mode"
    deletion_strategy = f"{headers_prefix}deletion-strategy"
    secret_store_token = f"{headers_prefix}secret-store-token"
    pipeline_arguments = f"{headers_prefix}pipeline-arguments"
    client_version = f"{headers_prefix}client-version"
    python_version = f"{headers_prefix}client-python-version"
    backend_version = f"{headers_prefix}be-version"
    ui_version = f"{headers_prefix}ui-version"
    ui_clear_cache = f"{headers_prefix}ui-clear-cache"


class FeatureStorePartitionByField(mlrun.common.types.StrEnum):
    name = "name"  # Supported for feature-store objects

    def to_partition_by_db_field(self, db_cls):
        if self.value == FeatureStorePartitionByField.name:
            return db_cls.name
        else:
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"Unknown group by field: {self.value}"
            )


class RunPartitionByField(mlrun.common.types.StrEnum):
    name = "name"  # Supported for runs objects
    project_and_name = "project_and_name"  # Supported for runs objects

    def to_partition_by_db_field(self, db_cls):
        if self.value == RunPartitionByField.name:
            return db_cls.name
        elif self.value == RunPartitionByField.project_and_name:
            return db_cls.project, db_cls.name
        else:
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"Unknown group by field: {self.value}"
            )


class ArtifactPartitionByField(mlrun.common.types.StrEnum):
    name = "name"  # Supported for artifacts objects
    project_and_name = "project_and_name"  # Supported for artifacts objects

    def to_partition_by_db_field(self, db_cls):
        if self.value == ArtifactPartitionByField.name:
            return db_cls.key
        elif self.value == ArtifactPartitionByField.project_and_name:
            return db_cls.project, db_cls.key
        else:
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"Unknown group by field: {self.value}"
            )


class SortField(mlrun.common.types.StrEnum):
    created = "created"
    updated = "updated"

    def to_db_field(self, db_cls):
        if self.value == SortField.created:
            # not doing type check to prevent import that will cause a cycle
            if db_cls.__name__ == "Run":
                return db_cls.start_time
            return db_cls.created
        elif self.value == SortField.updated:
            return db_cls.updated
        else:
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"Unknown sort by field: {self.value}"
            )


class OrderType(mlrun.common.types.StrEnum):
    asc = "asc"
    desc = "desc"

    def to_order_by_predicate(self, db_field):
        if self.value == OrderType.asc:
            return db_field.asc()
        else:
            return db_field.desc()


labels_prefix = "mlrun/"


class LabelNames:
    schedule_name = f"{labels_prefix}schedule-name"


class APIStates:
    online = "online"
    waiting_for_migrations = "waiting_for_migrations"
    migrations_in_progress = "migrations_in_progress"
    migrations_failed = "migrations_failed"
    migrations_completed = "migrations_completed"
    offline = "offline"
    waiting_for_chief = "waiting_for_chief"

    @staticmethod
    def terminal_states():
        return [APIStates.online, APIStates.offline]

    @staticmethod
    def description(state: str):
        return {
            APIStates.online: "API is online",
            APIStates.waiting_for_migrations: "API is waiting for migrations to be triggered. "
            "Send POST request to /api/operations/migrations to trigger it",
            APIStates.migrations_in_progress: "Migrations are in progress",
            APIStates.migrations_failed: "Migrations failed, API can't be started",
            APIStates.migrations_completed: "Migrations completed, API is waiting to become online",
            APIStates.offline: "API is offline",
            APIStates.waiting_for_chief: "API is waiting for chief to be ready",
        }.get(state, f"Unknown API state '{state}'")


class ClusterizationRole:
    chief = "chief"
    worker = "worker"


class LogsCollectorMode:
    legacy = "legacy"
    sidecar = "sidecar"
    best_effort = "best-effort"



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd
import plotly.graph_objects as go

from mlrun import get_or_create_ctx
from mlrun.artifacts import PlotlyArtifact


def my_job(context, p1=1, p2="x"):
    # load MLRUN runtime context (will be set by the runtime framework e.g. KubeFlow)

    # get parameters from the runtime context (or use defaults)

    # access input metadata, values, files, and secrets (passwords)
    print(f"Run: {context.name} (uid={context.uid})")
    print(f"Params: p1={p1}, p2={p2}")
    access_key = context.get_secret("ACCESS_KEY")
    if access_key:
        print("Access key retrieved successfully.")
    input_file = context.get_input("infile.txt", "infile.txt").get()
    print(f"File\n{input_file}\n")

    # Run some useful code e.g. ML training, data prep, etc.

    # log scalar result values (job result metrics)
    context.log_result("accuracy", p1 * 2)
    context.log_result("loss", p1 * 3)
    context.set_label("framework", "sklearn")

    # log various types of artifacts (file, web page, table), will be versioned and visible in the UI
    context.log_artifact(
        "model",
        body=b"abc is 123",
        local_path="model.txt",
        labels={"framework": "xgboost"},
    )
    context.log_artifact(
        "html_result", body=b"<b> Some HTML <b>", local_path="result.html"
    )

    # create a plotly output (will show in the pipelines UI)
    x = np.arange(10)
    fig = go.Figure(data=go.Scatter(x=x, y=x**2))

    # Create a PlotlyArtifact using the figure and log it
    plotly_artifact = PlotlyArtifact(figure=fig, key="plotly")
    context.log_artifact(plotly_artifact)

    raw_data = {
        "first_name": ["Jason", "Molly", "Tina", "Jake", "Amy"],
        "last_name": ["Miller", "Jacobson", "Ali", "Milner", "Cooze"],
        "age": [42, 52, 36, 24, 73],
        "testScore": [25, 94, 57, 62, 70],
    }
    df = pd.DataFrame(raw_data, columns=["first_name", "last_name", "age", "testScore"])
    context.log_dataset("mydf", df=df, stats=True)


if __name__ == "__main__":
    context = get_or_create_ctx("train", project="test-project")
    p1 = context.get_param("p1", 1)
    p2 = context.get_param("p2", "a-string")
    my_job(context, p1, p2)



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable, Optional, Union

import numpy as np
import tensorflow as tf
from tensorflow import keras
from tensorflow.python.keras.callbacks import Callback

import mlrun

from ..._common import LoggingMode
from ..._dl_common.loggers import Logger
from ..utils import TFKerasTypes, is_keras_3


class LoggingCallback(Callback):
    """
    Callback for collecting data during training / evaluation. All the collected data will be available in this callback
    post the training / evaluation process and can be accessed via the 'training_results', 'validation_results',
    'static_hyperparameters', 'dynamic_hyperparameters' and 'summaries' properties.
    """

    def __init__(
        self,
        context: mlrun.MLClientCtx = None,
        dynamic_hyperparameters: Optional[
            dict[
                str,
                Union[list[Union[str, int]], Callable[[], TFKerasTypes.TrackableType]],
            ]
        ] = None,
        static_hyperparameters: Optional[
            dict[str, Union[TFKerasTypes.TrackableType, list[Union[str, int]]]]
        ] = None,
        auto_log: bool = False,
    ):
        """
        Initialize a logging callback with the given hyperparameters and logging configurations.

        :param context:                 MLRun context to automatically log its parameters if 'auto_log' is True.
        :param dynamic_hyperparameters: If needed to track a hyperparameter dynamically (sample it each epoch) it should
                                        be passed here. The parameter expects a dictionary where the keys are the
                                        hyperparameter chosen names and the values are a key chain from the model. A key
                                        chain is a list of keys and indices to know how to access the needed
                                        hyperparameter from the model. If the hyperparameter is not of accessible from
                                        the model, a custom callable method can be passed. For example, to track the
                                        'lr' attribute of an optimizer and a custom parameter, one should pass:
                                        {
                                            "learning rate": ["optimizer", "lr"],
                                            "custom_parameter": get_custom_parameter
                                        }
        :param static_hyperparameters:  If needed to track a hyperparameter one time per run it should be passed here.
                                        The parameter expects a dictionary where the keys are the
                                        hyperparameter chosen names and the values are the hyperparameter static value
                                        or a key chain just like the dynamic hyperparameter. For example, to track the
                                        'epochs' of an experiment run, one should pass:
                                        {
                                            "epochs": 7
                                        }
        :param auto_log:                Whether to enable auto logging, trying to track common static and dynamic
                                        hyperparameters.
        """
        super().__init__()
        self._supports_tf_logs = True

        # Store the configurations:
        self._dynamic_hyperparameters_keys = (
            dynamic_hyperparameters if dynamic_hyperparameters is not None else {}
        )
        self._static_hyperparameters_keys = (
            static_hyperparameters if static_hyperparameters is not None else {}
        )

        # Initialize the logger:
        self._logger = Logger(context=context)

        # For calculating the batch's values we need to collect the epochs sums:
        # [Metric: str] -> [Sum: float]
        self._training_epoch_sums = {}  # type: Dict[str, float]
        self._validation_epoch_sums = {}  # type: Dict[str, float]

        # Setup the flags:
        self._is_training = None  # type: bool
        self._auto_log = auto_log

    def get_training_results(self) -> dict[str, list[list[float]]]:
        """
        Get the training results logged. The results will be stored in a dictionary where each key is the metric name
        and the value is a list of lists of values. The first list is by epoch and the second list is by iteration
        (batch).

        :return: The training results.
        """
        return self._logger.training_results

    def get_validation_results(self) -> dict[str, list[list[float]]]:
        """
        Get the validation results logged. The results will be stored in a dictionary where each key is the metric name
        and the value is a list of lists of values. The first list is by epoch and the second list is by iteration
        (batch).

        :return: The validation results.
        """
        return self._logger.validation_results

    def get_training_summaries(self) -> dict[str, list[float]]:
        """
        Get the training summaries of the metrics results. The summaries will be stored in a dictionary where each key
        is the metric names and the value is a list of all the summary values per epoch.

        :return: The training summaries.
        """
        return self._logger.training_summaries

    def get_validation_summaries(self) -> dict[str, list[float]]:
        """
        Get the validation summaries of the metrics results. The summaries will be stored in a dictionary where each key
        is the metric names and the value is a list of all the summary values per epoch.

        :return: The validation summaries.
        """
        return self._logger.validation_summaries

    def get_static_hyperparameters(self) -> dict[str, TFKerasTypes.TrackableType]:
        """
        Get the static hyperparameters logged. The hyperparameters will be stored in a dictionary where each key is the
        hyperparameter name and the value is his logged value.

        :return: The static hyperparameters.
        """
        return self._logger.static_hyperparameters

    def get_dynamic_hyperparameters(
        self,
    ) -> dict[str, list[TFKerasTypes.TrackableType]]:
        """
        Get the dynamic hyperparameters logged. The hyperparameters will be stored in a dictionary where each key is the
        hyperparameter name and the value is a list of his logged values per epoch.

        :return: The dynamic hyperparameters.
        """
        return self._logger.dynamic_hyperparameters

    def get_epochs(self) -> int:
        """
        Get the overall epochs this callback participated in.

        :return: The overall epochs this callback participated in.
        """
        return self._logger.epochs

    def get_training_iterations(self) -> int:
        """
        Get the overall training iterations this callback participated in.

        :return: The overall training iterations this callback participated in.
        """
        return self._logger.training_iterations

    def get_validation_iterations(self) -> int:
        """
        Get the overall validation iterations this callback participated in.

        :return: The overall validation iterations this callback participated in.
        """
        return self._logger.validation_iterations

    def on_train_begin(self, logs: Optional[dict] = None):
        """
        Called once at the beginning of training process (one time call).

        :param logs: Currently no data is passed to this argument for this method but that may change in the
                     future.
        """
        self._is_training = True
        self._setup_run()

    def on_test_begin(self, logs: Optional[dict] = None):
        """
        Called at the beginning of evaluation or validation. Will be called on each epoch according to the validation
        per epoch configuration.

        :param logs: Currently no data is passed to this argument for this method but that may change in the
                     future.
        """
        # Check if needed to mark this run as evaluation:
        if self._is_training is None:
            self._is_training = False
            self._logger.set_mode(mode=LoggingMode.EVALUATION)

        # If this callback is part of evaluation and not training, need to check if the run was setup:
        if not self._is_training:
            self._setup_run()

    def on_test_end(self, logs: Optional[dict] = None):
        """
        Called at the end of evaluation or validation. Will be called on each epoch according to the validation
        per epoch configuration. The recent evaluation / validation results will be summarized and logged.

        :param logs: Currently no data is passed to this argument for this method but that may change in the
                     future.
        """
        # Store the metrics average of this epoch:
        for metric_name, epoch_values in self._logger.validation_results.items():
            # Check if needed to initialize:
            if metric_name not in self._logger.validation_summaries:
                self._logger.validation_summaries[metric_name] = []
            self._logger.log_validation_summary(
                metric_name=metric_name,
                result=float(sum(epoch_values[-1]) / len(epoch_values[-1])),
            )

    def on_epoch_begin(self, epoch: int, logs: Optional[dict] = None):
        """
        Called at the start of an epoch, logging it and appending a new epoch to the logger's dictionaries.

        :param epoch: Integer, index of epoch.
        :param logs:  Currently no data is passed to this argument for this method but that may change in the
                      future.
        """
        # Log a new epoch:
        self._logger.log_epoch()

        # Reset the metrics sum:
        for sum_dictionary in [self._training_epoch_sums, self._validation_epoch_sums]:
            for metric in sum_dictionary:
                sum_dictionary[metric] = 0

    def on_epoch_end(self, epoch: int, logs: Optional[dict] = None):
        """
        Called at the end of an epoch, logging the training summaries and the current dynamic hyperparameters values.

        :param epoch: Integer, index of epoch.
        :param logs:  Metric results for this training epoch, and for the validation epoch if validation is
                      performed. Validation result keys are prefixed with `val_`. For training epoch, the values of the
                      `Model`'s metrics are returned. Example : `{'loss': 0.2, 'acc': 0.7}`.
        """
        # Store the last training result of this epoch:
        for metric_name, epoch_values in self._logger.training_results.items():
            # Check if needed to initialize:
            if metric_name not in self._logger.training_summaries:
                self._logger.training_summaries[metric_name] = []
            self._logger.log_training_summary(
                metric_name=metric_name, result=float(epoch_values[-1][-1])
            )

        # Update the dynamic hyperparameters dictionary:
        if self._dynamic_hyperparameters_keys:
            for name, key_chain in self._dynamic_hyperparameters_keys.items():
                self._logger.log_dynamic_hyperparameter(
                    parameter_name=name,
                    value=self._get_hyperparameter(key_chain=key_chain),
                )

    def on_train_batch_begin(self, batch: int, logs: Optional[dict] = None):
        """
        Called at the beginning of a training batch in `fit` methods. The logger will check if this batch is needed to
        be logged according to the configuration. Note that if the `steps_per_execution` argument to `compile` in
        `tf.keras.Model` is set to `N`, this method will only be called every `N` batches.

        :param batch: Integer, index of batch within the current epoch.
        :param logs:  Contains the return value of `model.train_step`. Typically, the values of the `Model`'s
                      metrics are returned. Example: `{'loss': 0.2, 'accuracy': 0.7}`.
        """
        self._logger.log_training_iteration()

    def on_train_batch_end(self, batch: int, logs: Optional[dict] = None):
        """
        Called at the end of a training batch in `fit` methods. The batch metrics results will be logged. Note that if
        the `steps_per_execution` argument to `compile` in `tf.keras.Model` is set to `N`, this method will only be
        called every `N` batches.

        :param batch: Integer, index of batch within the current epoch.
        :param logs:  Aggregated metric results up until this batch.
        """
        self._on_batch_end(
            results_dictionary=self._logger.training_results,
            sum_dictionary=self._training_epoch_sums,
            logs=logs,
        )

    def on_test_batch_begin(self, batch: int, logs: Optional[dict] = None):
        """
        Called at the beginning of a batch in `evaluate` methods. Also called at the beginning of a validation batch in
        the `fit` methods, if validation data is provided. The logger will check if this batch is needed to be logged
        according to the configuration. Note that if the `steps_per_execution` argument to `compile` in `tf.keras.Model`
        is set to `N`, this method will only be called every `N` batches.

        :param batch: Integer, index of batch within the current epoch.
        :param logs:  Contains the return value of `model.test_step`. Typically, the values of the `Model`'s
                      metrics are returned.  Example: `{'loss': 0.2, 'accuracy': 0.7}`.
        """
        self._logger.log_validation_iteration()

    def on_test_batch_end(self, batch: int, logs: Optional[dict] = None):
        """
        Called at the end of a batch in `evaluate` methods. Also called at the end of a validation batch in the `fit`
        methods, if validation data is provided. The batch metrics results will be logged. Note that if the
        `steps_per_execution` argument to `compile` in `tf.keras.Model` is set to `N`, this method will only be called
        every `N` batches.

        :param batch: Integer, index of batch within the current epoch.
        :param logs:  Aggregated metric results up until this batch.
        """
        self._on_batch_end(
            results_dictionary=self._logger.validation_results,
            sum_dictionary=self._validation_epoch_sums,
            logs=logs,
        )

    def _setup_run(self):
        """
        After the trainer / evaluator run begins, this method will be called to setup the logger, logging all the
        hyperparameters pre run (epoch 0).
        """
        # Setup the hyperparameters dictionaries:
        if self._auto_log:
            self._add_auto_hyperparameters()

        # Static hyperparameters:
        for name, value in self._static_hyperparameters_keys.items():
            if isinstance(value, list):
                # Its a parameter that needed to be extracted via key chain.
                self._logger.log_static_hyperparameter(
                    parameter_name=name,
                    value=self._get_hyperparameter(key_chain=value),
                )
            else:
                # Its a custom hyperparameter.
                self._logger.log_static_hyperparameter(parameter_name=name, value=value)

        # Dynamic hyperparameters:
        for name, key_chain in self._dynamic_hyperparameters_keys.items():
            self._logger.log_dynamic_hyperparameter(
                parameter_name=name,
                value=self._get_hyperparameter(key_chain=key_chain),
            )

    def _on_batch_end(self, results_dictionary: dict, sum_dictionary: dict, logs: dict):
        """
        Log the given metrics values to the given results dictionary.

        :param results_dictionary: One of 'self._logger.training_results' or 'self._logger.validation_results'.
        :param sum_dictionary:     One of 'self._training_epoch_sums' or 'self._validation_epoch_sums'.
        :param logs:               The loss and metrics results of the recent batch.
        """
        # Parse the metrics names in the logs:
        logs = self._parse_metrics(logs=logs)

        # Log the given metrics as needed:
        for metric_name, aggregated_value in logs.items():
            # Check if needed to initialize:
            if metric_name not in results_dictionary:
                results_dictionary[metric_name] = [[]]
                sum_dictionary[metric_name] = 0
            # Calculate the last value:
            elements_number = len(results_dictionary[metric_name][-1]) + 1
            elements_sum = sum_dictionary[metric_name]
            last_metric_value = elements_number * aggregated_value - elements_sum
            # Store the metric value at the current epoch:
            sum_dictionary[metric_name] += last_metric_value
            results_dictionary[metric_name][-1].append(float(last_metric_value))

    def _add_auto_hyperparameters(self):
        """
        Add auto log's hyperparameters if they are accessible. The automatic hyperparameters being added are:
        learning rate.  In addition to that, the context parameters (if available) will be logged as static
        hyperparameters as well.
        """
        # Log the context parameters:
        if self._logger.context is not None:
            self._logger.log_context_parameters()

        # Add learning rate:
        learning_rate_keys = [
            "learning_rate",
            "lr",
        ]  # "lr" is for backward compatibility in older keras versions.
        if all(
            learning_rate_key not in self._dynamic_hyperparameters_keys
            for learning_rate_key in learning_rate_keys
        ) and hasattr(self.model, "optimizer"):
            for learning_rate_key in learning_rate_keys:
                learning_rate_key_chain = ["optimizer", learning_rate_key]
                try:
                    self._get_hyperparameter(key_chain=learning_rate_key_chain)
                except (KeyError, IndexError, AttributeError, ValueError):
                    continue
                self._dynamic_hyperparameters_keys[learning_rate_key] = (
                    learning_rate_key_chain
                )
                break

    def _get_hyperparameter(
        self,
        key_chain: Union[
            Callable[[], TFKerasTypes.TrackableType], list[Union[str, int]]
        ],
    ) -> TFKerasTypes.TrackableType:
        """
        Access the hyperparameter from the model stored in this callback using the given key chain.

        :param key_chain: The keys and indices to get to the hyperparameter from the model or a callable method.

        :return: The hyperparameter value.

        :raise KeyError:   In case the one of the keys in the key chain is incorrect.
        :raise IndexError: In case the one of the keys in the key chain is incorrect.
        :raise MLRunInvalidArgumentError: In case the value is not trackable.
        """
        if isinstance(key_chain, Callable):
            # It is a custom callable method:
            value = key_chain()
        else:
            # Get the value using the provided key chain:
            value = self.model
            for key in key_chain:
                try:
                    if isinstance(key, int):
                        value = value[key]
                    else:
                        value = getattr(value, key)
                except KeyError or IndexError or AttributeError as KeyChainError:
                    raise KeyChainError(
                        f"Error during getting a hyperparameter value with the key chain {key_chain}. "
                        f"The {value.__class__} in it does not have the following key/index from the key provided: "
                        f"{key}"
                    )

        # Parse the value:
        if isinstance(value, (tf.Tensor, tf.Variable)) or (
            is_keras_3() and isinstance(value, (keras.KerasTensor, keras.Variable))
        ):
            if int(tf.size(value)) == 1:
                value = float(value)
            else:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    f"The parameter with the following key chain: {key_chain} is a tensorflow.Tensor with "
                    f"{value.numel()} elements. Tensorflow tensors are trackable only if they have 1 element."
                )
        elif isinstance(value, np.ndarray):
            if value.size == 1:
                value = float(value)
            else:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    f"The parameter with the following key chain: {key_chain} is a numpy.ndarray with {value.size} "
                    f"elements. numpy arrays are trackable only if they have 1 element."
                )
        elif not (isinstance(value, (float, int, str, bool))):
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"The parameter with the following key chain: {key_chain} is of type '{type(value)}'. The only "
                f"trackable types are: float, int, str and bool."
            )
        return value

    @staticmethod
    def _parse_metrics(logs: dict) -> dict:
        """
        Parse the default logs dictionary metrics names to be clean (without the 'val_' prefix).

        :param logs: The logs given from a callback method.

        :return: The parsed logs.
        """
        parsed_logs = {}
        for metric, value in logs.items():
            if metric.startswith("val_"):
                metric = metric[4:]
            parsed_logs[metric] = value

        return parsed_logs



# Copyright 2024 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import dataclasses
from typing import ClassVar, Optional

import numpy as np


@dataclasses.dataclass
class HistogramDistanceMetric(abc.ABC):
    """
    An abstract base class for distance metrics between histograms.

    :args distrib_t: array of distribution t (usually the latest dataset distribution)
    :args distrib_u: array of distribution u (usually the sample dataset distribution)

    Each distribution must contain nonnegative floats that sum up to 1.0.
    """

    distrib_t: np.ndarray
    distrib_u: np.ndarray

    NAME: ClassVar[str]

    # noinspection PyMethodOverriding
    def __init_subclass__(cls, *, metric_name: str, **kwargs) -> None:
        super().__init_subclass__(**kwargs)
        cls.NAME = metric_name

    @abc.abstractmethod
    def compute(self) -> float:
        raise NotImplementedError


class TotalVarianceDistance(HistogramDistanceMetric, metric_name="tvd"):
    """
    Provides a symmetric drift distance between two periods t and u
    Z - vector of random variables
    Pt - Probability distribution over time span t
    """

    def compute(self) -> float:
        """
        Calculate Total Variance distance.

        :returns:  Total Variance Distance.
        """
        return np.sum(np.abs(self.distrib_t - self.distrib_u)) / 2


class HellingerDistance(HistogramDistanceMetric, metric_name="hellinger"):
    """
    Hellinger distance is an f divergence measure, similar to the Kullback-Leibler (KL) divergence.
    It used to quantify the difference between two probability distributions.
    However, unlike KL Divergence the Hellinger divergence is symmetric and bounded over a probability space.
    The output range of Hellinger distance is [0,1]. The closer to 0, the more similar the two distributions.
    """

    def compute(self) -> float:
        """
        Calculate Hellinger Distance

        :returns: Hellinger Distance
        """
        return np.sqrt(
            max(
                1 - np.sum(np.sqrt(self.distrib_u * self.distrib_t)),
                0,  # numerical errors may produce small negative numbers, e.g. -1e-16.
                # However, Cauchy-Schwarz inequality assures this number is in the range [0, 1]
            )
        )


class KullbackLeiblerDivergence(HistogramDistanceMetric, metric_name="kld"):
    """
    KL Divergence (or relative entropy) is a measure of how one probability distribution differs from another.
    It is an asymmetric measure (thus it's not a metric) and it doesn't satisfy the triangle inequality.
    KL Divergence of 0, indicates two identical distributions.
    """

    @staticmethod
    def _calc_kl_div(
        actual_dist: np.ndarray, expected_dist: np.ndarray, zero_scaling: float
    ) -> float:
        """Return the asymmetric KL divergence"""
        # We take 0*log(0) == 0 for this calculation
        mask = actual_dist != 0
        actual_dist = actual_dist[mask]
        expected_dist = expected_dist[mask]
        with np.errstate(over="ignore"):
            # Ignore overflow warnings when dividing by small numbers,
            # resulting in inf:
            # RuntimeWarning: overflow encountered in true_divide
            relative_prob = actual_dist / np.where(
                expected_dist != 0, expected_dist, zero_scaling
            )
        return np.sum(actual_dist * np.log(relative_prob))

    def compute(
        self, capping: Optional[float] = None, zero_scaling: float = 1e-4
    ) -> float:
        """
        :param capping:      A bounded value for the KL Divergence. For infinite distance, the result is replaced with
                             the capping value which indicates a huge differences between the distributions.
        :param zero_scaling: Will be used to replace 0 values for executing the logarithmic operation.

        :returns: symmetric KL Divergence
        """
        t_u = self._calc_kl_div(self.distrib_t, self.distrib_u, zero_scaling)
        u_t = self._calc_kl_div(self.distrib_u, self.distrib_t, zero_scaling)
        result = t_u + u_t
        if capping and result == float("inf"):
            return capping
        return result



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shutil
import tarfile
import tempfile
import zipfile
from os import path, remove
from urllib.parse import urlparse

from git import Repo

import mlrun

from .helpers import logger


def _remove_directory_contents(target_dir):
    for filename in os.listdir(target_dir):
        file_path = os.path.join(target_dir, filename)
        if os.path.isfile(file_path) or os.path.islink(file_path):
            os.unlink(file_path)
        elif os.path.isdir(file_path):
            shutil.rmtree(file_path)


def _prep_dir(source, target_dir, suffix, secrets, clone):
    if not target_dir:
        raise ValueError("please specify a target (context) directory for clone")
    if clone and path.exists(target_dir) and path.isdir(target_dir):
        _remove_directory_contents(target_dir)

    temp_file = tempfile.NamedTemporaryFile(suffix=suffix, delete=False).name
    mlrun.get_dataitem(source, secrets).download(temp_file)
    return temp_file


def get_git_username_password_from_token(token):
    # Github's access tokens have a known prefix according to their type. See
    # https://docs.github.com/en/authentication/keeping-your-account-and-data-secure/about-authentication-to-github#githubs-token-formats
    # We distinguish new fine-grained access tokens (begin with "github_pat_" from classic tokens.
    if token.startswith("github_pat_") or token.startswith("glpat"):
        username = "oauth2"
        password = token
    else:
        username = token
        password = "x-oauth-basic"
    return username, password


def clone_zip(source, target_dir, secrets=None, clone=True):
    tmpfile = _prep_dir(source, target_dir, ".zip", secrets, clone)
    with zipfile.ZipFile(tmpfile, "r") as zf:
        zf.extractall(target_dir)
    remove(tmpfile)  # delete zipped file


def clone_tgz(source, target_dir, secrets=None, clone=True):
    tmpfile = _prep_dir(source, target_dir, ".tar.gz", secrets, clone)
    with tarfile.TarFile.open(tmpfile, "r:*") as tf:
        tf.extractall(path=target_dir)
    remove(tmpfile)  # delete zipped file


def get_repo_url(repo):
    url = ""
    remotes = [remote.url for remote in repo.remotes]
    if not remotes:
        return ""

    url = remotes[0]
    url = url.replace("https://", "git://")
    try:
        url = f"{url}#refs/heads/{repo.active_branch.name}"
    except Exception:
        pass

    return url


def add_credentials_git_remote_url(url: str, secrets=None) -> tuple[str, bool]:
    """Enrich a Git remote URL with credential related secrets, if any are available
    If no secrets are supplied, or if the secrets are insufficient, the original URL is returned
    Besides the URL, this function also returns a bool indicating if any enrichment was done

    :param url:     git remote URL to be enriched
    :param secrets: dict or SecretsStore with Git credentials e.g. secrets={"GIT_TOKEN": token}

    :returns: tuple with the final URL and a boolean indicating if any enrichment was done
    """

    def get_secret(key):
        return mlrun.get_secret_or_env(key, secret_provider=secrets)

    url_obj = urlparse(url)

    username = url_obj.username or get_secret("GIT_USERNAME") or get_secret("git_user")
    password = (
        url_obj.password
        or get_secret("GIT_PASSWORD")
        or get_secret("git_password")
        or ""
    )
    token = (
        get_secret("GITHUB_TOKEN")
        or get_secret("GITLAB_TOKEN")
        or get_secret("GIT_TOKEN")
    )
    if token:
        username, password = get_git_username_password_from_token(token)

    if username:
        return f"https://{username}:{password}@{url_obj.netloc}{url_obj.path}", True
    return url, False


def clone_git(url: str, context: str, secrets=None, clone: bool = True):
    """Clone a remote Git repository in the local context

    :param url:     git remote URL
    :param context: local directory in which the repository must be stored
    :param secrets: dict or SecretsStore with Git credentials e.g. secrets={"GIT_TOKEN": token}
    :param clone:   delete all files and folders in "context" if there are any
    """

    url_obj = urlparse(url)
    if not context:
        raise ValueError("please specify a target (context) directory for clone")

    if path.exists(context) and path.isdir(context):
        if clone:
            _remove_directory_contents(context)
        else:
            if os.path.exists(context) and len(os.listdir(context)) > 0:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    "Failed to load project from git, context directory is not empty. "
                    "Set clone param to True to remove the contents of the context directory."
                )
            try:
                repo = Repo(context)
                return get_repo_url(repo), repo
            except Exception:
                pass

    host = url_obj.hostname or "github.com"
    if url_obj.port:
        host += f":{url_obj.port}"

    clone_path = f"https://{host}{url_obj.path}"
    final_clone_path, is_path_enriched = add_credentials_git_remote_url(
        clone_path, secrets=secrets or {}
    )

    branch = None
    tag = None
    commit = None
    if url_obj.fragment:
        refs = url_obj.fragment
        if refs.startswith("refs/heads/"):
            branch = refs.replace("refs/heads/", "")
        elif refs.startswith("refs/tags/"):
            tag = refs.replace("refs/tags/", "")
        elif refs.startswith("refs/commits/"):
            commit = refs.replace("refs/commits/", "")
        else:
            url = url.replace(f"#{refs}", f"#refs/heads/{refs}")
            branch = refs

    # when using the CLI and clone path was not enriched, username/password input will be requested via shell
    repo = Repo.clone_from(final_clone_path, context, single_branch=True, b=branch)

    if is_path_enriched:
        # override enriched clone path for security reasons
        repo.remotes[0].set_url(clone_path, final_clone_path)

    if tag_or_commit := tag or commit:
        repo.git.checkout(tag_or_commit)

    return url, repo


def extract_source(source: str, workdir=None, secrets=None, clone=True):
    if not source:
        return
    clone = clone if workdir else False
    target_dir = workdir or os.path.realpath("./code")
    if source.endswith(".zip"):
        clone_zip(source, target_dir, secrets, clone)
    elif source.endswith(".tar.gz"):
        clone_tgz(source, target_dir, secrets, clone)
    elif source.startswith("git://"):
        clone_git(source, target_dir, secrets, clone)
    else:
        if path.exists(source) and path.isdir(source):
            if workdir and workdir != source:
                raise ValueError("cannot specify both source and workdir")
            return path.realpath(source)
        raise ValueError(f"unsupported source format/path {source}")

    logger.info(f"extracting source from {source} to {target_dir}")
    return target_dir



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Union

import numpy as np
import pandas as pd

import mlrun

from .._common import LoggingMode
from .metric import Metric
from .utils import SKLearnUtils


class Estimator:
    """
    Class for handling metrics calculations during a run.
    """

    def __init__(
        self,
        context: mlrun.MLClientCtx = None,
        metrics: Optional[list[Metric]] = None,
    ):
        """
        Initialize an estimator with the given metrics. The estimator will log the calculated results using the given
        context.

        :param context: The context to log with.
        :param metrics: The metrics
        """
        # Store the context and metrics:
        self._context = context
        self._metrics = metrics if metrics is not None else []

        # Setup the logger's mode (default:  Training):
        self._mode = LoggingMode.TRAINING

        # Prepare the dictionaries to hold the results. Once they are logged they will be moved from one to another:
        self._logged_results = {}  # type: Dict[str, float]
        self._not_logged_results = {}  # type: Dict[str, float]

    @property
    def context(self) -> mlrun.MLClientCtx:
        """
        Get the logger's MLRun context.

        :return: The logger's MLRun context.
        """
        return self._context

    @property
    def results(self) -> dict[str, float]:
        """
        Get the logged results.

        :return: The logged results.
        """
        return self._logged_results

    def set_mode(self, mode: LoggingMode):
        """
        Set the estimator's mode.

        :param mode: The mode to set.
        """
        self._mode = mode

    def set_context(self, context: mlrun.MLClientCtx):
        """
        Set the context this logger will log with.

        :param context: The to be set MLRun context.
        """
        self._context = context

    def set_metrics(self, metrics: list[Metric]):
        """
        Update the metrics of this logger to the given list of metrics here.

        :param metrics: The list of metrics to override the current one.
        """
        self._metrics = metrics

    def is_probabilities_required(self) -> bool:
        """
        Check if probabilities are required in order to calculate some of the metrics.

        :return: True if probabilities are required by at least one metric and False otherwise.
        """
        return any(metric.need_probabilities for metric in self._metrics)

    def estimate(
        self,
        y_true: Union[np.ndarray, pd.DataFrame, pd.Series],
        y_pred: Union[np.ndarray, pd.DataFrame, pd.Series],
        is_probabilities: bool = False,
    ):
        """
        Calculate the results according to the 'is_probabilities' flag and log them.

        :param y_true:           The ground truth values to send for the metrics functions.
        :param y_pred:           The predictions to send for the metrics functions.
        :param is_probabilities: True if the 'y_pred' is a prediction of probabilities (from 'predict_proba') and False
                                 if not. Default: False.
        """
        # Calculate the metrics results:
        self._calculate_results(
            y_true=y_true, y_pred=y_pred, is_probabilities=is_probabilities
        )

        # Log if a context is available:
        if self._context is not None:
            # Log the results in queue:
            self._log_results()
            # Commit:
            self._context.commit(completed=False)

    def _calculate_results(
        self,
        y_true: Union[np.ndarray, pd.DataFrame, pd.Series],
        y_pred: Union[np.ndarray, pd.DataFrame, pd.Series],
        is_probabilities: bool,
    ):
        """
        Calculate the results from all the metrics in the estimator.

        :param y_true:           The ground truth values to send for the metrics functions.
        :param y_pred:           The predictions to send for the metrics functions.
        :param is_probabilities: True if the 'y_pred' is a prediction of probabilities (from 'predict_proba') and False
                                 if not.
        """
        # Use squeeze to remove redundant dimensions:
        y_true = np.squeeze(SKLearnUtils.to_array(dataset=y_true))
        y_pred = np.squeeze(SKLearnUtils.to_array(dataset=y_pred))

        # Calculate the metrics:
        for metric in self._metrics:
            if metric.need_probabilities == is_probabilities:
                self._not_logged_results[metric.name] = metric(y_true, y_pred)

        # Add evaluation prefix if in Evaluation mode:
        if self._mode == LoggingMode.EVALUATION:
            self._not_logged_results = {
                f"evaluation_{key}": value
                for key, value in self._not_logged_results.items()
            }

    def _log_results(self):
        """
        Log the calculated metrics results using the logger's context.
        """
        # Use the context to log each metric result:
        self._context.log_results(self._not_logged_results)

        # Collect the logged results:
        self._logged_results = {**self._logged_results, **self._not_logged_results}

        # Clean the not logged results dictionary:
        self._not_logged_results = {}



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC
from typing import Optional, Union

import mlrun
from mlrun.artifacts import Artifact
from mlrun.datastore import is_store_uri
from mlrun.features import Feature

from .._common import ModelHandler
from .utils import MLTypes, MLUtils


class MLModelHandler(ModelHandler, ABC):
    """
    Abstract class for a machine learning framework model handling, enabling loading, saving and logging it during runs.
    """

    def __init__(
        self,
        model: MLTypes.ModelType = None,
        model_path: MLTypes.PathType = None,
        model_name: Optional[str] = None,
        modules_map: Union[
            dict[str, Union[None, str, list[str]]], MLTypes.PathType
        ] = None,
        custom_objects_map: Union[
            dict[str, Union[str, list[str]]], MLTypes.PathType
        ] = None,
        custom_objects_directory: MLTypes.PathType = None,
        context: mlrun.MLClientCtx = None,
        **kwargs,
    ):
        """
        Initialize the handler. The model can be set here so it won't require loading. Note you must provide at least
        one of 'model' and 'model_path'. If a model is not given, the files in the model path will be collected
        automatically to be ready for loading.

        :param model:                    Model to handle or None in case a loading parameters were supplied.
        :param model_path:               Path to the directory with the model files. Can be passed as a model object
                                         path in the following format:
                                         'store://models/<PROJECT_NAME>/<MODEL_NAME>:<VERSION>'
        :param model_name:               The model name for saving and logging the model:
                                         * Mandatory for loading the model from a local path.
                                         * If given a logged model (store model path) it will be read from the artifact.
                                         * If given a loaded model object and the model name is None, the name will be
                                           set to the model's object name / class.
        :param modules_map:              A dictionary of all the modules required for loading the model. Each key
                                         is a path to a module and its value is the object name to import from it. All
                                         the modules will be imported globally. If multiple objects needed to be
                                         imported from the same module a list can be given. The map can be passed as a
                                         path to a json file as well. For example:
                                         {
                                             "module1": None,  # => import module1
                                             "module2": ["func1", "func2"],  # => from module2 import func1, func2
                                             "module3.sub_module": "func3",  # => from module3.sub_module import func3
                                         }
                                         If the model path given is of a store object, the modules map will be read from
                                         the logged modules map artifact of the model.
        :param custom_objects_map:       A dictionary of all the custom objects required for loading the model. Each key
                                         is a path to a python file and its value is the custom object name to import
                                         from it. If multiple objects needed to be imported from the same py file a list
                                         can be given. The map can be passed as a path to a json file as well. For
                                         example:
                                         {
                                             "/.../custom_optimizer.py": "optimizer",
                                             "/.../custom_layers.py": ["layer1", "layer2"]
                                         }
                                         All the paths will be accessed from the given 'custom_objects_directory',
                                         meaning each py file will be read from 'custom_objects_directory/<MAP VALUE>'.
                                         If the model path given is of a store object, the custom objects map will be
                                         read from the logged custom object map artifact of the model.
                                         Notice: The custom objects will be imported in the order they came in this
                                         dictionary (or json). If a custom object is depended on another, make sure to
                                         put it below the one it relies on.
        :param custom_objects_directory: Path to the directory with all the python files required for the custom
                                         objects. Can be passed as a zip file as well (will be extracted during the run
                                         before loading the model). If the model path given is of a store object, the
                                         custom objects files will be read from the logged custom object artifact of the
                                         model.
        :param context:                  MLRun context to work with for logging the model.
        :param model_format:             The format to use for saving and loading the model. Should be passed as a
                                         member of the class 'ModelFormats'. Default: 'ModelFormats.PKL'.

        :raise MLRunInvalidArgumentError: In case one of the given parameters are invalid.
        """
        # Setup additional properties for logging a ml model into a ModelArtifact:
        self._algorithm = None  # type: str
        self._sample_set = None  # type: MLTypes.DatasetType
        self._target_columns = None  # type: MLTypes.TargetColumnsNamesType
        self._feature_vector = None  # type: str
        self._feature_weights = None  # type: List[float]

        # Continue the initialization:
        super().__init__(
            model=model,
            model_path=model_path,
            model_name=model_name,
            modules_map=modules_map,
            custom_objects_map=custom_objects_map,
            custom_objects_directory=custom_objects_directory,
            context=context,
            **kwargs,
        )

    @property
    def algorithm(self) -> str:
        """
        Get the model algorithm set in this handler.

        :return: The handler's model algorithm.
        """
        return self._algorithm

    @property
    def sample_set(self) -> MLTypes.DatasetType:
        """
        Get the sample dataset set in this handler.

        :return: The handler's sample dataset.
        """
        return self._sample_set

    @property
    def target_columns(self) -> MLTypes.TargetColumnsNamesType:
        """
        Get the sample dataset target columns set in this handler.

        :return: The handler's sample dataset target columns names.
        """
        return self._target_columns

    @property
    def feature_vector(self) -> str:
        """
        Get the feature vector set in this handler.

        :return: The handler's feature vector.
        """
        return self._feature_vector

    @property
    def feature_weights(self) -> list[float]:
        """
        Get the feature weights set in this handler.

        :return: The handler's feature weights.
        """
        return self._feature_weights

    def set_algorithm(self, algorithm: str):
        """
        Set the algorithm this model will be logged with.

        :param algorithm: The algorithm to set.
        """
        self._algorithm = algorithm

    def set_sample_set(
        self, sample_set: Union[MLTypes.DatasetType, mlrun.DataItem, str]
    ):
        """
        Set the sample set this model will be logged with. The sample set will be casted to a pd.DataFrame. Can be sent
        as a DataItem and as a store object string.

        :param sample_set: The sample set to set.

        :raise MLRunInvalidArgumentError: In case the sample set store uri is incorrect or if the sample set type is not
                                          supported.
        """
        # Parse the store uri to a DataItem:
        if isinstance(sample_set, str):
            if not is_store_uri(sample_set):
                raise mlrun.errors.MLRunInvalidArgumentError(
                    f"The provided sample set string is an invalid store uri: {sample_set}"
                )
            sample_set = mlrun.get_dataitem(sample_set)

        # Parse the DataItem to a DataFrame:
        if isinstance(sample_set, mlrun.DataItem):
            sample_set = sample_set.as_df()

        # Set the sample set casting it to a DataFrame:
        self._sample_set = MLUtils.to_dataframe(dataset=sample_set)

    def set_target_columns(self, target_columns: MLTypes.TargetColumnsNamesType):
        """
        Set the ground truth column names of the sample set this model will be logged with.

        :param target_columns: The ground truth (y) columns names to set.
        """
        self._target_columns = target_columns

    def set_feature_vector(self, feature_vector: str):
        """
        Set the feature vector this model will be logged with.

        :param feature_vector: The feature store feature vector uri to set
                               (store://feature-vectors/<project>/<name>[:tag]).
        """
        self._feature_vector = feature_vector

    def set_feature_weights(self, feature_weights: list[float]):
        """
        Set the feature weights this model will be logged with.

        :param feature_weights: The feature weights to set, one per input column.
        """
        self._feature_weights = feature_weights

    def log(
        self,
        tag: str = "",
        labels: Optional[dict[str, Union[str, int, float]]] = None,
        parameters: Optional[dict[str, Union[str, int, float]]] = None,
        inputs: Optional[list[Feature]] = None,
        outputs: Optional[list[Feature]] = None,
        metrics: Optional[dict[str, Union[int, float]]] = None,
        artifacts: Optional[dict[str, Artifact]] = None,
        extra_data: Optional[dict[str, MLTypes.ExtraDataType]] = None,
        algorithm: Optional[str] = None,
        sample_set: MLTypes.DatasetType = None,
        target_columns: MLTypes.TargetColumnsNamesType = None,
        feature_vector: Optional[str] = None,
        feature_weights: Optional[list[float]] = None,
    ):
        """
        Log the model held by this handler into the MLRun context provided.

        :param tag:             Tag of a version to give to the logged model. Will override the stored tag in this
                                handler.
        :param labels:          Labels to log the model with. Will be joined to the labels set.
        :param parameters:      Parameters to log with the model. Will be joined to the parameters set.
        :param inputs:          A list of features this model expects to receive - the model's input ports. If already
                                set, will be overridden by the inputs given here.
        :param outputs:         A list of features this model expects to return - the model's output ports. If already
                                set, will be overridden by the outputs given here.
        :param metrics:         Metrics results to log with the model.
        :param artifacts:       Artifacts to log the model with. Will be joined to the registered artifacts and added to
                                the extra data.
        :param extra_data:      Extra data to log with the model.
        :param algorithm:       The algorithm of this model. If None it will be read as the model's class name.
        :param sample_set:      Sample set to use for getting the model's inputs, outputs and base stats for model
                                monitoring. Do not pass both sample set and inputs / outputs.
        :param target_columns:  The ground truth (y) labels names.
        :param feature_vector:  Feature store feature vector uri (store://feature-vectors/<project>/<name>[:tag])
        :param feature_weights: List of feature weights, one per input column.

        :raise MLRunRuntimeError:         In case is no model in this handler.
        :raise MLRunInvalidArgumentError: If a context is missing.
        """
        # Update the algorithm (set default if both values are None):
        if algorithm is not None:
            self._algorithm = algorithm
        elif self._algorithm is None:
            self.set_algorithm(algorithm=self._model.__class__.__name__)

        # Update the sample set:
        if sample_set is not None:
            self.set_sample_set(sample_set=sample_set)
        if target_columns is not None:
            self.set_target_columns(target_columns=target_columns)

        # Update the feature parameters:
        if feature_vector is not None:
            self.set_feature_vector(feature_vector=feature_vector)
        if feature_weights is not None:
            self.set_feature_weights(feature_weights=feature_weights)

        # Continue with the handler logging:
        super().log(
            tag=tag,
            labels=labels,
            parameters=parameters,
            inputs=inputs,
            outputs=outputs,
            metrics=metrics,
            artifacts=artifacts,
            extra_data=extra_data,
            algorithm=self._algorithm,
            sample_set=self._sample_set,
            target_columns=self._target_columns,
            feature_vector=self._feature_vector,
            feature_weights=self._feature_weights,
        )

    def update(
        self,
        labels: Optional[dict[str, Union[str, int, float]]] = None,
        parameters: Optional[dict[str, Union[str, int, float]]] = None,
        inputs: Optional[list[Feature]] = None,
        outputs: Optional[list[Feature]] = None,
        metrics: Optional[dict[str, Union[int, float]]] = None,
        artifacts: Optional[dict[str, Artifact]] = None,
        extra_data: Optional[dict[str, MLTypes.ExtraDataType]] = None,
        feature_vector: Optional[str] = None,
        feature_weights: Optional[list[float]] = None,
    ):
        """
        Update the model held by this handler into the MLRun context provided, updating the model's artifact properties
        in the same model path provided.

        :param labels:          Labels to update or add to the model.
        :param parameters:      Parameters to update or add to the model.
        :param inputs:          A list of features this model expects to receive - the model's input ports. If already
                                set, will be overridden by the inputs given here.
        :param outputs:         A list of features this model expects to return - the model's output ports. If already
                                set, will be overridden by the outputs given here.
        :param metrics:         Metrics results to log with the model.
        :param artifacts:       Artifacts to update or add to the model. Will be joined to the registered artifacts and
                                added to the extra data.
        :param extra_data:      Extra data to update or add to the model.
        :param feature_vector:  Feature store feature vector uri (store://feature-vectors/<project>/<name>[:tag])
        :param feature_weights: List of feature weights, one per input column.

        :raise MLRunInvalidArgumentError: In case a context is missing or the model path in this handler is missing or
                                          not of a store object.
        """
        # Update the feature parameters:
        if feature_vector is not None:
            self._feature_vector = feature_vector
        if feature_weights is not None:
            self._feature_weights = feature_weights

        # Continue with the handler update:
        super().update(
            labels=labels,
            parameters=parameters,
            inputs=inputs,
            outputs=outputs,
            metrics=metrics,
            artifacts=artifacts,
            extra_data=extra_data,
            feature_vector=self._feature_vector,
            feature_weights=self._feature_weights,
        )



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import tensorflow as tf
from packaging import version
from tensorflow import keras

import mlrun
from mlrun.data_types import ValueType

from .._dl_common import DLTypes, DLUtils


class TFKerasTypes(DLTypes):
    """
    Typing hints for the TensorFlow.Keras framework.
    """

    # Every model in tf.keras must inherit from tf.keras.Model:
    ModelType = keras.Model


class TFKerasUtils(DLUtils):
    """
    Utilities functions for the TensorFlow.Keras framework.
    """

    @staticmethod
    def convert_value_type_to_tf_dtype(
        value_type: str,
    ) -> tf.DType:
        """
        Get the 'tensorflow.DType' equivalent to the given MLRun value type.

        :param value_type: The MLRun value type to convert to tensorflow data type.

        :return: The 'tensorflow.DType' equivalent to the given MLRun data type.

        :raise MLRunInvalidArgumentError: If tensorflow is not supporting the given data type.
        """
        # Initialize the mlrun to tensorflow data type conversion map:
        conversion_map = {
            ValueType.BOOL: tf.bool,
            ValueType.INT8: tf.int8,
            ValueType.INT16: tf.int16,
            ValueType.INT32: tf.int32,
            ValueType.INT64: tf.int64,
            ValueType.UINT8: tf.uint8,
            ValueType.UINT16: tf.uint16,
            ValueType.UINT32: tf.uint32,
            ValueType.UINT64: tf.uint64,
            ValueType.BFLOAT16: tf.bfloat16,
            ValueType.FLOAT16: tf.float16,
            ValueType.FLOAT: tf.float32,
            ValueType.DOUBLE: tf.float64,
        }

        # Convert and return:
        if value_type in conversion_map:
            return conversion_map[value_type]
        raise mlrun.errors.MLRunInvalidArgumentError(
            f"The ValueType given is not supported in tensorflow: '{value_type}'."
        )

    @staticmethod
    def convert_tf_dtype_to_value_type(
        tf_dtype: tf.DType,
    ) -> str:
        """
        Convert the given tensorflow data type to MLRun data type. All the CUDA supported data types are supported.
        For more information regarding tensorflow data types,
        visit: https://www.tensorflow.org/api_docs/python/tf/dtypes

        :param tf_dtype: The tensorflow data type to convert to MLRun's data type. Expected to be a 'tensorflow.dtype'
                         or 'str'.

        :return: The MLRun value type converted from the given data type.

        :raise MLRunInvalidArgumentError: If the tensorflow data type is not supported by MLRun.
        """
        # Initialize the tensorflow to mlrun data type conversion map:
        conversion_map = {
            tf.bool.name: ValueType.BOOL,
            tf.int8.name: ValueType.INT8,
            tf.int16.name: ValueType.INT16,
            tf.int32.name: ValueType.INT32,
            tf.int64.name: ValueType.INT64,
            tf.uint8.name: ValueType.UINT8,
            tf.uint16.name: ValueType.UINT16,
            tf.uint32.name: ValueType.UINT32,
            tf.uint64.name: ValueType.UINT64,
            tf.bfloat16.name: ValueType.BFLOAT16,
            tf.half.name: ValueType.FLOAT16,
            tf.float16.name: ValueType.FLOAT16,
            tf.float32.name: ValueType.FLOAT,
            tf.double.name: ValueType.DOUBLE,
            tf.float64.name: ValueType.DOUBLE,
        }

        # Parse the given tensorflow data type to string:
        if isinstance(tf_dtype, tf.DType):
            tf_dtype = tf_dtype.name

        # Convert and return:
        if tf_dtype in conversion_map:
            return conversion_map[tf_dtype]
        raise mlrun.errors.MLRunInvalidArgumentError(
            f"MLRun value type is not supporting the given tensorflow data type: '{tf_dtype}'."
        )


def is_keras_3() -> bool:
    """
    Check if the current Keras version is 3.x.

    :return: True if Keras version is 3.x, False otherwise.
    """
    return hasattr(keras, "__version__") and version.parse(
        keras.__version__
    ) >= version.parse("3.0.0")



# Copyright 2024 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from datetime import datetime

import mlrun.feature_store.steps
from mlrun.common.schemas.model_monitoring import (
    EventFieldType,
    EventKeyMetrics,
)
from mlrun.utils import logger


class ProcessBeforeTDEngine(mlrun.feature_store.steps.MapClass):
    def __init__(self, **kwargs):
        """
        Process the data before writing to TDEngine. This step create the relevant keys for the TDEngine table,
        including project name, custom metrics, time column, and table name column.

        :returns: Event as a dictionary which will be written into the TDEngine Predictions table.
        """
        super().__init__(**kwargs)

    def do(self, event):
        event[EventFieldType.PROJECT] = event[EventFieldType.FUNCTION_URI].split("/")[0]
        event[EventKeyMetrics.CUSTOM_METRICS] = json.dumps(
            event.get(EventFieldType.METRICS, {})
        )
        event[EventFieldType.TIME] = event.get(EventFieldType.TIMESTAMP)
        event[EventFieldType.TABLE_COLUMN] = "_" + event.get(EventFieldType.ENDPOINT_ID)

        return event


class ErrorExtractor(mlrun.feature_store.steps.MapClass):
    def __init__(self, **kwargs):
        """
        Prepare the event for insertion into the TDEngine error table
        """
        super().__init__(**kwargs)

    def do(self, event):
        error = str(event.get("error"))
        if len(error) > 1000:
            error = error[-1000:]
            logger.warning(
                f"Error message exceeds 1000 chars: The error message writen to TSDB will be it last "
                f"1000 chars, Error:  {error}",
                event=event,
            )
        timestamp = datetime.fromisoformat(event.get("when"))
        endpoint_id = event[EventFieldType.ENDPOINT_ID]
        event = {
            EventFieldType.MODEL_ERROR: error,
            EventFieldType.ERROR_TYPE: EventFieldType.INFER_ERROR,
            EventFieldType.ENDPOINT_ID: endpoint_id,
            EventFieldType.TIME: timestamp,
            EventFieldType.PROJECT: event[EventFieldType.FUNCTION_URI].split("/")[0],
            EventFieldType.TABLE_COLUMN: "_err_"
            + event.get(EventFieldType.ENDPOINT_ID),
        }
        logger.info("Write error to errors TSDB table", event=event)
        return event
